/* 
   BARCODE.PRG
   NOTE: This is provided "as is" to the dUFLP library
         This depends on the true type font: UPCDN___.TTF
                    This font must be installed
                    on the users' system along with
                    anything else you need to do,
                    and InstallShield Express will
                    not handle adding fonts ...
         Make sure for any report you wish to generate the
         barcode for that you set the font to this font.
         
   NOTE: For dUFLP 2021, the true type font (UPCDN___.TTF) is
   stored in the folder "BarCode". You still need to install it
   to Windows ... 
*/

/*************************************************************************
*  PROGRAM:        Barcode(lnNumber)
*
*  Programmer:     Richard Forrest - Freedom Graphics, Lytton, B.C., Canada
*
*  Last Modified:  07/98 - Visual dBase V7.01 conversion
*			(commenting and slight code modification required to
*			 backdate it to V5.6)
*  Revisions:	   01/97 - Visual dBase V5.6 version
*
*  VERSION:        Visual dBASE 7.01
*
*  DESCRIPTION:
*
*  Code to generate a UPC-D barcode string from an item number.
*
*  UPC-D is listed as a generic UPC code used for in-house labeling: i.e. it
*  is used where a business wants to establish a barcode system that is used
*  to track things within the business, but may (and probably will) not be
*  of any value outside the business. This is great for retail in-house tagging.
*
*  It does not require a manufacturer number, and is supposedly unlimited
*  as to length. It wants only numeric input - 0 to 9 - no other character
*  values.
*
*  I have restricted this routine to only the odd values, and have not
*  experimented with splitting even values. The checksum portion of this
*  code is not effected by odd or even values, but I do not know which side
*  of the center guard to put the extra digit on, and have not bothered to
*  figure that out, as I have no need of this right now.
*
*  USAGE:         barcode(lnNumber)
*
*  where lnNumber is the value you want to generate a barcode for. It can
*  be either an integer or a character value, the code will check.
*
*  This routine is intended for use with my UPCD font, and has character
*  mappings appropriate for that font. i.e. - it is pretty useless unless you
*  install the included UPC-D font and use that to display and print the
*  generated bar code.
*  
*  BTW - I generally just include this code into a cc file of business rules
*  for the app I am working on.
**************************************************************************/

PROCEDURE BARCODE(lnNumber)

// Declare variables

LOCAL lnItemNumber, lcItemNumber, lcBarCode, lnOdd, lnEven, lcCheck,;
      lnTemp, lnCheck, lnLength, lnLength2, llFlipFlop 

      lnNumber2 = lnNumber // Make sure that the datatype is defined

// Initialize variables
lnOdd = 0      && Counter to sum the odd numbers for a checksum
lnEven = 0     && Counter to sum the even numbers for a checksum
llFlipFlop = true   && A logical variable to indicate or even numbers
lcLeft = ""    && A character string to hold the left half of the barcode
lcRight = ""   && A character string to hold the right half of the barcode

// Check for valid inputs

IF EMPTY(lnNumber2) // If this is empty - no barcode!
   MSGBOX("You must supply an odd length number!","BarCode Routine WARNING",16)
   RETURN 0
ENDIF

// Change the item number to a string and get the length
IF TYPE("lnNumber2") # "C"  // Check for character or numeric
   lnItemNumber = INT(lnNumber2)
   lcItemNumber = STR(lnItemNumber,20)
ELSE  // It is already a character value
   lcItemNumber = lnNumber2
ENDIF

// Now that we have a character value, we make sure that it is trimed
lcItemNumber = LTRIM(RTRIM(lcItemNumber))

lnLength =  LEN(lcItemNumber) // Get the length of the number
lnLength2 = lnLength

// Test to make sure that we only have numeric characters ...
DO WHILE lnLength2 >0
   lcTestChar = SUBSTR(lcItemNumber,lnLength2,1)
   IF lcTestChar < "0" OR lcTestChar > "9"
      MSGBOX("Item number " + lcItemNumber +" is not valid for UPC-D code!";
      			,"Barcode Routine WARNING",16)
      RETURN 0
   ENDIF
   lnLength2 --
ENDDO

IF MOD(lnLength,2) = 0 // If mod = 0 the item is even length - no barcode!
   MSGBOX(lcItemNumber +" is even - you must use an odd number of digits!";
   				,"Barcode Routine WARNING",16)
   RETURN 0
ENDIF

/* 
* Parse the ItemNumber to get the odd and even totals. This is
* necessary to calculate the checksum, which is the total of
* the odd digits plus 3 times the total of the even digits
* modulus 10 subtracted from 10. The rightmost digit is odd.
*/

DO WHILE lnLength > 0
   lnTemp = VAL(SUBSTR(lcItemNumber,lnLength,1))
   IF llFlipFlop
      lnEven = lnEven + lnTemp
      llFlipFlop = false
   ELSE
      lnOdd = lnOdd + lnTemp
      llFlipFlop = true
   ENDIF
   lnLength --
ENDDO

// Get the checksum value

lnCheck = 10 - (MOD(((lnEven * 3)+lnOdd),10))

// Check for the case where lnCheck is 10 - make it 0 instead
IF lnCheck = 10
   lnCheck = 0
ENDIF

// Change it to a string value with a length of 1 character

lcCheck = LTRIM(RTRIM(STR(lnCheck)))
/**************************************************************************
* Add the checksum to the barcode string
**************************************************************************/
lcBarCode = lcItemNumber + lcCheck

/**************************************************************************
* Divide the lcBarCode string in half, and convert the second half to
* right hand characters (coded A through J in the UPC-D font)
**************************************************************************/
   
   lnLength2 = LEN(lcBarCode)
   lnLength3 = lnLength2/2
   // Getting the left side is easy...
   lcLeft = SUBSTR(lcBarCode,1,lnLength3)
   // Getting the right side is a little more difficult
   DO WHILE lnLength2 > lnLength3
      lcBCodeChar = SUBSTR(lcBarCode,lnLength2--,1)
      IF lcBCodeChar == "0"    // This could be a DO CASE structure instead
         lcRightConvert = "A"
      ELSE IF lcBCodeChar == "1"
         lcRightConvert = "B"
      ELSE IF lcBCodeChar == "2"
         lcRightConvert = "C"
      ELSE IF lcBCodeChar == "3"
         lcRightConvert = "D"
      ELSE IF lcBCodeChar == "4"
         lcRightConvert = "E"
      ELSE IF lcBCodeChar == "5"
         lcRightConvert = "F"
      ELSE IF lcBCodeChar == "6"
         lcRightConvert = "G"
      ELSE IF lcBCodeChar == "7"
         lcRightConvert = "H"
      ELSE IF lcBCodeChar == "8"
         lcRightConvert = "I"
      ELSE IF lcBCodeChar == "9"
         lcRightConvert = "J"
      ELSE 
      // There should not be any other possibilities, but ...
         MSGBOX("Bar code generation error!","Barcode Routine WARNING",16)
      ENDIF
      lcRight  = lcRightConvert + lcRight
   ENDDO

   /* We now have a left string consisting of the first half of the numbers and a
   * right string consisting of the second half plus the checksum, but converted
   * to letters. We now put a left bracket in front, a star in the center and a
   * right bracket on the end. This is just because that is where I stored the
   * appropriate guard characters in my UPC-D barcode font.                   
   */
   
   lcBarCode = "("+lcLeft+"*"+lcRight+")"
   
   /* Now we return the barcode to the calling routine. It can be displayed or
   * printed using the UPC-D font from any form, or even with streaming output
   * if you declare the appropriate font. I have had no trouble with printing
   * barcodes at size 18 on a 300 dpi laser printer onto common label stock.
   */
   
   RETURN lcBarCode
