/*
   -----------------------------------------------------------------------
   Programmer..: Ken Mayer
   Date........: February 15, 2007
   Notes.......: Compares the file structure of two tables.
                 This will check for:
                    TableType being the same. If not, we don't do anything else.
                    Fields, checking field information (fieldname, type,
                       etc).
                    Indexes, checking index information (indexName, expression,
                       etc.)
   Written for.: dBPlus
   Rev. History: Original Version
   Calls.......: None
   Called by...: Any
   Usage.......: CompareStructure( <cTable1>, <cTable2> [, <bOutFile>] )
   Examples....: ? CompareStructure( "MyTable1", "MyTable2" )
                 ? CompareStructure( ":MyDatabase:MyTable1", "MyTable2" )
                 ? CompareStructure( ":MyDatabase:MyTable1", ":MyDatabase:MyTable2" )
                 ? CompareStructure( ":MyDatabase1:MyTable1", ":MyDatabase2:MyTable2" )
                 ? CompareStructure( "MyTable1", ":MyDatabase:MyTable2" )

                 The difference in these different versions is whether or
                 not a database alias was passed to reference the table
                 in any case. Note that the alias must be used as "normal"
                 in dBASE, with the colon (:) delimiter. 

                 To create an output file if errors occur, use the third
                 parameter:
                 ? CompareStructure( "MyTable1", ":MyDatabase:MyTable2", true )

   Returns.....: logical -- true means tables match, false means they don't.
                    If option to output to a file is used, true will
                    simply erase the file, as there is no output to be
                    concerned with, and false will call a simple editor
                    style form to display the output.
   Parameters..: cTable1  = First table to compare
                 cTable2  = Second table to compare
                 bOutFile = logical -- true means to use one, false means
                            to not.
   ---------------------------------------------------------------------
*/
function CompareStructure

   parameters cTable1, cTable2, bOutFile

   // error handling
   if pCount() < 2
      msgbox( "Please pass two table names to compare!",;
              "ERROR!", 16 )
      return
   endif

   // default the output file option to false
   if pCount() == 2
      bOutFile = false
   endif

   local cData1, cData2, cTableName1, cTableName2
   local oData1, oData2, oTD1, oTD2
   local bMatch, oOutFile
   bMatch = true

   // if we need an output file:
   if bOutFile
      if file( "CompStru.txt" )
         erase CompStru.txt
      endif
      oOutFile = new file()
      oOutFile.create( "CompStru.txt" )
      oOutFile.puts( "Compare Table Structures" )
      oOutFile.puts( cTable1+" and "+cTable2 )
      oOutFile.puts( ""+date() )
      oOutFile.puts( replicate( "-", 80 ) )
   endif

   // The first thing we need to do is determine if we have
   // a database reference for either of these tables.
   // If we do, we we need to extract the table name
   // from the database, and use these separately.
   cData1 = ""
   cTableName1 = ""
   if at( ":", cTable1 ) == 1 and at( ":", cTable1, 2 ) > 0
      // we have the proper delimiters for a database alias:
      cData1      = substr( cTable1, 2, at( ":", cTable1, 2 ) -2 )
      cTableName1 = right( cTable1, len( cTable1 ) - at( ":", cTable1, 2 ) )
   else
      cTableName1 = cTable1
   endif

   cData2 = ""
   cTableName2 = ""
   if at( ":", cTable2 ) == 1 and at( ":", cTable2, 2 ) > 0
      // we have the proper delimiters for a database alias:
      cData2      = substr( cTable2, 2, at( ":", cTable2, 2 ) -2 )
      cTableName2 = right( cTable2, len( cTable2 ) - at( ":", cTable2, 2 ) )
   else
      cTableName2 = cTable2
   endif

   // set up database objects if needed:
   try
      if not empty( cData1 )
         oData1 = new database()
         oData1.databaseName := cData1
         oData1.active := true
      endif
      if not empty( cData2 )
         oData2 = new database()
         oData2.databaseName := cData2
         oData2.active := true
      endif
   catch( Exception E )
      if bOutFile
         oOutFile.puts( "Error "+e.code+" -- "+e.message )
      endif // bOutFile
      bMatch := false
   endtry

   // set up the tableDef objects:
   if bMatch == true
      try
         oTD1 = new TableDef()
         if not empty( cData1 )
            oTD1.database := oData1
         endif
         oTD1.tableName := cTableName1
         oTD1.load()

         oTD2 = new TableDef()
         if not empty( cData2 )
            oTD2.database := oData2
         endif
         oTD2.tableName := cTableName2
         oTD2.load()
      catch( Exception E )
         if bOutFile
            oOutFile.puts( "Error "+e.code+" -- "+e.message )
         endif // bOutFile
         bMatch := false
      endtry
   endif // bMatch == true

   // check to see if the table types match:
   if bMatch == true
      if oTD1.tableType # oTD2.tableType
         if bOutFile
            oOutFile.puts( "Table Types do not match!"+chr(13)+chr(10)+;
                           "Table "+cTableName1+" is "+oTD1.tableType+"."+chr(13)+chr(10)+;
                           "Table "+cTableName2+" is "+oTD2.tableType+"."+chr(13) )
         endif // bOutFile
         bMatch := false
      endif // comparison
   endif // bMatch == true

   // ----------------------------------------------------------
   // now we start the fun part ... first, we only do this
   // if we have reached this point without error, because
   // the other errors are pretty much fatal to this process:
   if bMatch == true
      // first, the simplest test of all --
      // check to see if the number of fields
      // is the same:
      if oTD1.fields.size # oTD2.fields.size
         if bOutFile
            oOutFile.puts( "Number of fields does not match!"+chr(13)+chr(10)+;
                           "Table "+cTableName1+" has "+oTD1.fields.size+" fields."+chr(13)+chr(10)+;
                           "Table "+cTableName2+" has "+oTD2.fields.size+" fields."+chr(13) )
         endif // bOutFile
         bMatch := false
      endif // check number of fields

      local i
      // move on ... test each field
      for i = 1 to oTD1.fields.size
          // if we don't have any more fields in table 2,
          // comparison is pointless, eh?
          if oTD2.fields.size => i
             if bOutFile
                oOutFile.puts( replicate("-", 30 ) )
                oOutFile.puts( "Comparing Field "+int(i)+":" )
             endif

             // loop through fields and compare fieldName, type, 
             // length, and if numeric, decimalLength.
             if oTD1.fields[ i ].fieldName # oTD2.fields[ i ].fieldName
                if bOutFile
                   oOutFile.puts( "FieldNames do not match!"+chr(13)+chr(10)+;
                                  "Table "+cTableName1+" Field "+i+" fieldname is: "+;
                                          oTD1.fields[i].fieldName+"."+chr(13)+chr(10)+;
                                  "Table "+cTableName2+" Field "+i+" fieldname is: "+;
                                          oTD2.fields[i].fieldName+"."+chr(13)+chr(10) )
                endif // bOutFile
                bMatch := false
             endif // fieldName comparison

             // if we have a lock field, we don't want to try comparing
             // the type, length, etc. ... The field is defined differently.
             if oTD1.fields[ i ].fieldName # "_DBASELOCK"
                // continue with type:
                if oTD1.fields[ i ].type # oTD2.fields[ i ].type
                   if bOutFile
                      oOutFile.puts( "Field Types do not match!"+chr(13)+chr(10)+;
                                     "Table "+cTableName1+" Field "+i+" type is: "+;
                                          oTD1.fields[i].type+"."+chr(13)+chr(10)+;
                                 "Table "+cTableName2+" Field "+i+" type is: "+;
                                          oTD2.fields[i].type+"."+chr(13)+chr(10) )
                   endif // bOutFile
                   bMatch := false
                endif // type comparison

                // continue with length
                if oTD1.fields[ i ].length # oTD2.fields[ i ].length
                   if bOutFile
                      oOutFile.puts( "Field Lengths/Widths do not match!"+chr(13)+chr(10)+;
                                     "Table "+cTableName1+" Field "+i+" length is: "+;
                                             int( oTD1.fields[i].length )+"."+chr(13)+chr(10)+;
                                     "Table "+cTableName2+" Field "+i+" length is: "+;
                                             int( oTD2.fields[i].length )+"."+chr(13)+chr(10) )
                   endif // bOutFile
                   bMatch := false
                endif // compare length

                // continue with decimalLength
                if oTD1.fields[ i ].type == "NUMERIC" or;
                   oTD1.fields[ i ].type == "FLOAT" or;
                   oTD1.fields[ i ].type == "NUMBER"
                   if oTD1.fields[ i ].decimalLength # oTD2.fields[ i ].decimalLength
                      if bOutFile
                         oOutFile.puts( "Field Decimals do not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Field "+i+" decimal length is: "+;
                                                 int( oTD1.fields[i].decimalLength )+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Field "+i+" decimal length is: "+;
                                                 int( oTD2.fields[i].decimalLength )+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // compare decimalLength
                endif // numeric

             endif // _dBASELOCK field

          endif // number of fields in table 2
      next // i

      // ------------------------------------------------------------------
      // and now to do something similar, but for the 
      // index tags. For this we need to know what the
      // table type is, because the properties
      // are different if we have a .DBF versus
      // other index tags:
      if oTD1.tableType == "DBASE"
         if oTD1.indexes.size # oTD2.indexes.size
            if bOutFile
               oOutFile.puts( replicate("-", 60 ) )
               oOutFile.puts( "Comparing Indexes:" )
               oOutFile.puts( "Number of index tags does not match!"+chr(13)+chr(10)+;
                              "Table "+cTableName1+" has "+oTD1.indexes.size+" index tags."+chr(13)+chr(10)+;
                              "Table "+cTableName2+" has "+oTD2.indexes.size+" index tags."+chr(13) )
            endif // bOutFile
            bMatch := false
         endif // indexes.size

         // now we loop through the indexes in table1, and compare, but
         // only if we have any index tags to compare:
         if oTD1.indexes.size > 0 and oTD2.indexes.size > 0

            // loop through fields and compare indexName, descending, 
            // expression, forExpression, unique.
            for i = 1 to oTD1.indexes.size
                // check to make sure we have an index tag to compare
                // against ...
                if oTD2.indexes.size => i

                   if bOutFile
                      oOutFile.puts( replicate("-", 30 ) )
                      oOutFile.puts( "Comparing Index "+int(i)+":" )
                   endif

                   if oTD1.indexes[ i ].indexName # oTD2.indexes[ i ].indexName
                      if bOutFile
                         oOutFile.puts( "IndexNames do not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" indexName is: "+;
                                                oTD1.indexes[i].indexName+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" indexName is: "+;
                                                oTD2.indexes[i].indexName+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // indexName comparison

                   if oTD1.indexes[ i ].expression # oTD2.indexes[ i ].expression
                      if bOutFile
                         oOutFile.puts( "Expression does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" expression is: "+;
                                                oTD1.indexes[i].expression+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" expression is: "+;
                                                oTD2.indexes[i].expression+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // expression comparison

                   if oTD1.indexes[ i ].forExpression # oTD2.indexes[ i ].forExpression
                      if bOutFile
                         oOutFile.puts( "ForExpression does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" forExpression is: "+;
                                                oTD1.indexes[i].forExpression+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" forExpression is: "+;
                                                oTD2.indexes[i].forExpression+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // forExpression comparison

                   if oTD1.indexes[ i ].descending # oTD2.indexes[ i ].descending
                      if bOutFile
                         oOutFile.puts( "Descending does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" descending is: "+;
                                                oTD1.indexes[i].descending+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" descending is: "+;
                                                oTD2.indexes[i].descending+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // descending comparison

                   if oTD1.indexes[ i ].unique # oTD2.indexes[ i ].unique
                      if bOutFile
                         oOutFile.puts( "Unique does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" unique is: "+;
                                                oTD1.indexes[i].unique+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" unique is: "+;
                                                oTD2.indexes[i].unique+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // descending comparison

                endif // number of indexes in table 2 ...
            next // i
         endif // we have any indexes at all ...

      else // we're dealing with .DB or SQL Server indexes
         
         if oTD1.indexes.size # oTD2.indexes.size
            if bOutFile
               oOutFile.puts( replicate("-", 60 ) )
               oOutFile.puts( "Comparing Indexes:" )
               oOutFile.puts( "Number of index tags does not match!"+chr(13)+chr(10)+;
                              "Table "+cTableName1+" has "+oTD1.indexes.size+" index tags."+chr(13)+chr(10)+;
                              "Table "+cTableName2+" has "+oTD2.indexes.size+" index tags."+chr(13) )
            endif // bOutFile
            bMatch := false
         endif

         // now we loop through the indexes in table1, and compare, but
         // only if we have any index tags to compare:
         if oTD1.indexes.size > 0 and oTD2.indexes.size > 0

            // loop through fields and compare indexName, descending, 
            // expression, forExpression, unique.
            for i = 1 to oTD1.indexes.size
                // check to make sure we have an index tag to compare
                // against ...
                if oTD2.indexes.size => i

                   if bOutFile
                      oOutFile.puts( replicate("-", 30 ) )
                      oOutFile.puts( "Comparing Index "+int(i)+":" )
                   endif

                   if oTD1.indexes[ i ].indexName # oTD2.indexes[ i ].indexName
                      if bOutFile
                         oOutFile.puts( "IndexNames do not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" indexName is: "+;
                                                oTD1.indexes[i].indexName+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" indexName is: "+;
                                                oTD2.indexes[i].indexName+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // indexName comparison

                   if oTD1.indexes[ i ].fields # oTD2.indexes[ i ].fields
                      if bOutFile
                         oOutFile.puts( "Fields do not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" fields is: "+;
                                                oTD1.indexes[i].fields+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" fields is: "+;
                                                oTD2.indexes[i].fields+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // fields comparison

                   if oTD1.indexes[ i ].caseSensitive # oTD2.indexes[ i ].caseSensitive
                      if bOutFile
                         oOutFile.puts( "CaseSensitive does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" caseSensitive is: "+;
                                                oTD1.indexes[i].caseSensitive+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" caseSensitive is: "+;
                                                oTD2.indexes[i].caseSensitive+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // caseSensitive comparison

                   if oTD1.indexes[ i ].descending # oTD2.indexes[ i ].descending
                      if bOutFile
                         oOutFile.puts( "Descending does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" descending is: "+;
                                                oTD1.indexes[i].descending+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" descending is: "+;
                                                oTD2.indexes[i].descending+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // descending comparison

                   if oTD1.indexes[ i ].unique # oTD2.indexes[ i ].unique
                      if bOutFile
                         oOutFile.puts( "Unique does not match!"+chr(13)+chr(10)+;
                                        "Table "+cTableName1+" Index "+i+" unique is: "+;
                                                oTD1.indexes[i].unique+"."+chr(13)+chr(10)+;
                                        "Table "+cTableName2+" Index "+i+" unique is: "+;
                                                oTD2.indexes[i].unique+"."+chr(13)+chr(10) )
                      endif // bOutFile
                      bMatch := false
                   endif // descending comparison

                endif // number of indexes in table 2 ...

            next // i

         endif // we have any indexes to work wtih
      endif // table type for indexes ...

   endif // bMatch == true

   // cleanup:
   if type( "oTD1" ) # "U"
      release object oTD1
   endif
   if type( "oTD2" ) # "U"
      release object oTD2
   endif
   if not empty( cData1 ) and type( "oData1" ) # "U"
      oData1.active := false
      release object oData1
      oData1 = null
   endif
   if not empty( cData2 ) and type( "oData2" ) # "U"
      oData2.active := false
      release object oData2
      oData2 = null
   endif

   // display the output, if we have anything:
   if bOutFile 
      if not bMatch
         oOutFile.puts( replicate( "-", 80 ) )
         oOutFile.puts( "*** These tables do not have the same structure. ***" )
      endif

      // close and release:
      oOutFile.close()
      release object oOutFile

      // display errors
      if not bMatch // we have at least one error:
         local oF
         oF = new CompStruForm()
         oF.readModal()
         release object oF

         // ask if they want to erase the file:
         if msgbox( "If you select 'No', the file 'CompStru.txt"+chr(13)+;
                    "will be left on the hard drive. If you select"+chr(13)+;
                    "'Yes', it will be erased.",;
                    "Erase Output File?", 32+4+256 ) == 6 // Yes
            erase CompStru.txt
         endif
      else // we have no errors, don't keep the output
           // file:
         erase CompStru.txt
      endif
   endif // bOutFile

return bMatch
// End of program CompareStructure.prg

// ---------------------------------------------------------------------------
// Form for display of errors ...
class CompStruForm of FORM
   with (this)
      metric = 6	// Pixels
      height = 485.0
      left = 162.0
      top = 8.0
      width = 675.0
      text = "Table Structure Comparison Report"
      autoCenter = true
      mdi = false
      maximize = false
      minimize = false
   endwith

   this.TITLETEXT = new TEXT(this)
   with (this.TITLETEXT)
      height = 29.0
      left = 0.0
      top = 0.0
      width = 675.0
      anchor = 2	// Top
      colorNormal = "Highlight/BtnFace"
      alignHorizontal = 1	// Center
      fontSize = 18.0
      text = "Table Structure Comparison Report"
   endwith

   this.ERRORFILEED = new EDITOR(this)
   with (this.ERRORFILEED)
      height = 414.0
      left = 3.0
      top = 32.0
      width = 671.0
      modify = false
      fontName = "Courier New"
      dataLink = "FILE CompStru.txt"
   endwith

   this.CLOSEBTN = new PUSHBUTTON(this)
   with (this.CLOSEBTN)
      onClick = {; form.close()}
      height = 31.0
      left = 254.0
      top = 450.0
      width = 83.0
      text = "Close"
      upBitmap = "RESOURCE #1005"
   endwith

   this.PRINTBTN = new PUSHBUTTON(this)
   with (this.PRINTBTN)
      onClick = class::PRINTBTN_ONCLICK
      height = 31.0
      left = 338.0
      top = 450.0
      width = 83.0
      text = "Print"
      upBitmap = "RESOURCE #857"
   endwith

   function PRINTBTN_onClick
      // first, get the printer:
      if choosePrinter()
         set printer to
         set printer on
         // open a file object in read-only mode
         local oFile
         oFile = new file()
         oFile.open( "CompStru.txt", "R" )
         // loop through the file one line at a time:
         do while not oFile.eof()
            // output it:
            ? oFile.gets()
         enddo
         // cleanup
         oFile.close()
         release object oFile
         // close the printer and send
         // report ...
         set printer off
         set printer to
         close printer
      endif
   return

endclass
// end of form used for this program.
