// -- File: STACK.CC

CLASS Stack OF Array()
/*
    -----------------------------------------------------------------
    Description:
 
       This custom class is a sub-class of the built-in Visual dBASE 
       Array class. The original methods were written for dBASE IV by 
       Frank A. Deviney, Jr., but were completely rewritten for 
       Visual dBASE. (A couple of Frank's routines were released as 
       unnecessary)
 
       A stack is a single-column array. Elements are added by 
       "pushing" them, they are removed by "popping" them. 
 
       Note: Because a stack is a subclass of array, all properties 
       and methods of arrays (size, scan, etc.) might be available. 
       HOWEVER, as of 12/31/1998, this class has all of these
       properties and methods protected, and they are not
       available for use outside of the class itself ...
 
    Programmers: 
 
       Ken Mayer
 
    History: 
 
       11/27/1996 -- original version of these routines as a class.
       12/31/1998 -- PROTECTed the array properties/methods
                     making them unavailable for use outside
                     of the class's methods ...

                     You can check this by doing the following:
                     set procedure to stack.cc additive
                     oMyStack = new Stack()
                     inspect( oMyStack )
                     // you should not see ANY of the
                     // properties except for className,
                     // or any of the methods except
                     // those defined below ...
 
    Instantiation: 
 
       set procedure to stack.cc additive
       oMyStack = new Stack()
 
    Methods:
       More details on each of these is contained in the commentary 
       for each method ...
    
       Release()
         Releases this object (and others) from memory
  
       Push()
         Add element to stack
 
       Pop()
         Remove element from stack
 
       GetTop()
         Return top element, but does not remove it
  
       Empty()
         Returns logical true/false depending on whether or not
         the stack is empty.
 
    Properties:
       None
   ------------------------------------------------------------------
*/

   protect dimensions, size, ;
           add, delete, dir, dirExt, element, fields,;
           fill, grow, insert, resize, scan, sort, subscript

   FUNCTION Release
   /*
      ---------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/24/1996
       Notes.......: Used to release this object from memory.
       Written for.: Visual dBASE 5.5
       Rev. History: 11/24/1996 -- Original Code
       Calls.......: none
       Usage.......: stack.Release()
       Example.....: oMyStack.Release()
       Returns.....: Null
       Parameters..: None
      ---------------------------------------------------------------
   */

      release object this

   RETURN NULL
   // -- EoM: Release()

   FUNCTION Empty
   /*
      ---------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/04/1996
       Notes.......: Checks if stack is empty.
       Written for.: Visual dBASE 5.5a
       Rev. History: 11/04/1996 -- Original
       Calls.......: None
       Usage.......: oMyStack.empty()
       Example.....: lError = oMyStack.empty()
       Returns.....: logical: true if stack is empty, false otherwise
       Parameters..: none
      ---------------------------------------------------------------
   */
   
   RETURN ( this.size == 0 )
   // -- EoM: Empty()

   FUNCTION Push ( xElement )
   /*
      ---------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/04/1996
       Notes.......: Adds an element to the stack. Stack elements can 
                     be of any type.
       Written for.: Visual dBASE
       Rev. History: 11/04/1996 -- Original
       Calls.......: None
       Usage.......: oMyStack.Push( <xElement> )
       Example.....: oMyStack.Push( cWord )
       Returns.....: None
       Parameters..: xElement = element to add to stack
      ---------------------------------------------------------------
   */

      // add a new element
      this.add( xElement )

   // -- EoM: Push()

   FUNCTION Pop
   /*
      ---------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/05/1996
       Notes.......: Remove top element from the stack (if stack not 
                       empty, if it is, it returns NULL).
       Written for.: Visual dBASE
       Rev. History: 11/05/1996 -- Original
       Calls.......: stack::Empty()
       Usage.......: <stackName>.Pop()
       Example.....: xLastElement = oMyStack.Pop()
       Returns.....: Element being removed or NULL if stack is empty
                     (You can check for NULL: 
                                          if oMyStack.pop() = NULL)
       Parameters..: None
      ---------------------------------------------------------------
   */

      // if array is not empty:
      if not this.empty()
         // get "top" item (last one added)
         xReturn = this[ this.size ]
         // reduce the size of the array
         this.size--
      else
         // otherwise, return a NULL value
         xReturn = NULL
      endif

   RETURN xReturn
   // -- EoM: Pop()
   
   FUNCTION GetTop
   /*
      ---------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/05/1996
       Notes.......: Returns top element from the stack (if stack not 
                       empty) without 'popping' it.
       Written for.: Visual dBASE
       Rev. History: 11/05/1996 -- Original
       Calls.......: stack::Empty()
       Usage.......: <stackName>.GetTop()
       Example.....: xTopElement = oMyStack.GetTop()
       Returns.....: Topmost element in the stack, NULL if stack is 
                     empty (You can check for NULL: 
                            if oMyStack.pop() = NULL )
       Parameters..: None
      ---------------------------------------------------------------
   */
  
   RETURN ( iif( this.size > 0, this[ this.size ], NULL ) )
   // -- EoM: GetTop()
   
ENDCLASS
// -- End of Class: Stack

/* 
   ------------------------------------------------------------------
   End of File: STACK.CC
   ------------------------------------------------------------------
*/
