// -- File: CONVERT.CC

CLASS Convert
/*
   ----------------------------------------------------------------
   Description:

   CLASS Convert is a set of conversion routines, used to convert
   numbers to text, text to numbers, and a variety of other 
   conversions.

   Programmers:

    Ken Mayer
    Jay Parsons
    Kenneth Chan
    Angus Scott-Fleming
    Kelvin Smith
    Nick Carlin
    Oktay Amiry
    Erik Moeller 
    Gerardo Pasos
    Jean-Pierre Martel
    Mervyn Bick
    Gaetano De Luisi

    History: 
 
    11/24/1996    original version of these routines as a class.
    06/22/1998    Erik Moeller produced the new Num2GermanWords and
                  Thou2GermanWords methods ...
    06/23/1998    Erik sent me a fix for a minor problem with "und"
                  in Thou2GermanWords.
    10/20/1998    Minor updates
    November 29, 1999
                  Gerardo Pasos sent in the Mexican/Spanish versions
                  of Num2Words and Thou2Words ...
    July 9, 2001  R. Wingerter noticed that Thou2GermanWords
                  was returning incorrect values if using a Windows
                  driver, it's been updated for ANSI, but the original
                  is there in case you're using DOS (ASCII) drivers,
                  as Thou2GermanWordsASC.
    July 12, 2001 -- Same author as above modified code for
                  Thou2GermanWords to allow use with both
                  ASCII and ANSI language drivers. This removes
                  the need for Thou2GermanWordsASC().
    February, 2002 -- modified to use SetProc.prg rather than
                  dUFLP.H

    Spring, 2005 -- Added Thou2FrenchWords(), Num2FrenchWords()
                  (J.P. Martel)
    July, 2010   -- Modified Cash2Check() to handle non-US currencies.
    
    February, 2020 -- added Num2Word_ITA() and Thou2ITAWords() methods
        to convert numeric values to Italian text. (Maurizio S.)
        
    February, 2021 -- Updated Num2FrenchWords() -- Gaetano De Luisi,
        to deal with some incorrect translations and to update
        the format of numbers to comply with the spelling reform of 1990. 
        Also added Num2WordsBE() and Thou2BEWords() for Belgian.
                   
    Instantiation:
 
       set procedure to convert.cc additive
       oConvert = new Convert()
 
    Typical usage would be to call individual methods from here,
    such as:
 
      ? oConvert.Arabic( "XXIV" )

    **** NOTE: If you do not use the release() method of Convert,
         you will not be able to release the reference to the
         StringEX class that is created in the init() method.
 
    Methods:
      Note, more details are in the headers for each method, 
      including a listing of parameters, and what they mean, 
      as well as the complete history (as we know it) for each 
      method.
 
       Init()   
         handles any "initialization" necessary, is handled
         when a new instance of the Convert class is created.
 
       Release() 
         releases this class and the string class if it wasn't
         instantiated previously
 
       Arabic( <cRomanString> )
         Converts roman numeral (string) to arabic number
 
       BankRound()
         Rounds number according to banker's rules.
 
       Bin2Dec()
         Converts Binary numeral to decimal
 
       BreakLong() 
         Breaks out an unsigned integer held in four characters
 
       BreakWord()
         Break out an unsigned integer held in two characters
 
       ByteVal()
         Converts multi-byte value to decimal equivalent
 
       Cash2Check()
         Converts numeric to text, current version handles through
         a set of extra/optional parameters, currencies other than
         the U.S. dollars and cents.
 
       Cel2Fahr()
         Celsius to Fahrenheit
 
       CM2Inch()
         Centimeters to inches
 
       CUFT2Gal()
         Cubic feet to gallons
 
       Dec2Bin()
         Decimal value to binary equiv.
 
       Dec2Mkd() 
         Converts numeric value to eight-character string as IEEE 
         real
 
       Dec2Mki()
         Converts a small integer to a one-word equivalent
 
       Dec2Mkl()
         Converts integer to doubleword equivalent
 
       Dec2Mks()
         Converts numeric value to four-character string as IEEE 
         real
 
       Dec2MSks()
         Same as Dec2Mks but uses old MicroSoft format
 
       Dec2Oct()
         Decimal to Octal
 
       Deg2Num()
         Degrees to Numbers
 
       Fahr2Cel()
         Fahrenheit to Celcius
 
       Frac2Bin()
         Converts fractional part of decimal to binary
 
       Gal2CuFt()
         Gallons to Cubic Feet
 
       Gal2Ltr()
         Gallons to Liters
 
       Inch2CM()
         Inches to Centimeters
 
       KG2LB()
         Kilograms to pounds
 
       Km2Mile()
         Kilometers to Miles
 
       Km2Naut()
         Kilometers to Nautical Miles
 
       Lb2Kg()
         Pounds to Kilograms
 
       Long2Signed()
         Breaks signed integer out of four characters
 
       Ltr2Gal()
         Liters to Gallons
 
       MakeLong()
         Makes a four-character string from an unsigned integer
 
       MakeWord()
         Makes a two-character string from an unsigned integer
 
       Mile2KM()
         Miles to Kilometers
 
       Mkd2Dec()
         Converts IEEE long real to numeric (BASIC MKD$)
 
       Mki2Dec()
         Converts signed short integer (BASIC MKI$) to numeric
 
       Mkl2Dec()
         Converts signed long integer (BASIC MKL$) to numeric
 
       MSks2Dec()
         Converts old-style short real value (BASIC MKS$) to numeric
 
       Naut2KM()
         Nautical Miles to Kilometers
 
       Naut2Stat()
         Nautical Miles to Statute Miles
 
       Num2Real()
         Converts number to IEEE 4 or 8 byte real
 
       Num2Str()
         Converts number to string, like str(), but does not require
         number of decimals as parm.
 
       Num2Words()
         number to text
 
       NumVal()
         like val(), but only works on string that is all numeric
 
       Oct2Dec()
         Octal to decimal
 
       Ord()
         converts number to words, adding ending, such as th, nd, etc.
 
       Ordinal()
         same as Ord() for integers < 100
 
       Roman()
         Converts arabic to Roman Numerals
 
       Stat2Naut()
         Statute Miles to Nautical Miles

       Str2Float()
         Convert a string which contains a 4- or 8-byte IEEE
         floating point number to its actual numeric value.
 
       StuffLong()
         Stuffs an unisgned integer into four characters of a string
 
       StuffWord()
         Stuffs an unisgned integer into four characters of a string
 
       SU()
         Standard units
 
       Thou2Words()
         same for integers < 1000

       Num2GermanWords()
       Thou2GermanWords()
         These two routines are used to do the same thing
         that Num2Words() and Thou2Words() do for English --
         they return the German text ...

       Num2SpanishWords()
       Thou2SpanishWords()
         Same as above but for Mexican Spanish
         
         Num2Word_ITA()
         Thou2ITAWords()
            Same but for Italian
 
   ** Note that this cc file uses Source Aliasing, specifically the
   ** dUFLP source Alias. If you do not know how to set up
   ** a Source Alias, see the instructions in WHATS.NEW at
   ** the top of the file, or in README.TXT.
   ** Any files referenced by :dUFLP:filename  must be
   ** included in an executable if you are building 
   ** one ... make sure that your project includes these
   ** files.
   **
   ** This particular program uses:
   **           stringex.cc
   **           SetProc.prg
 
   ----------------------------------------------------------------
*/

   class::Init()   // initialize

   FUNCTION Init
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/09/1996
       Notes.......: Ensures that routines needed are available
                     It is required that this be used (see 
                     instructions above) for this class to 
                     function properly.
       Written for.: Visual dBASE 5.5
       Rev. History: 11/09/1996 -- Original Code
       Calls.......: string.cc
       Usage.......: convert.Init() // automatic when instantiated
       Example.....: _app.Convert.Init()
       Returns.....: Nul
       Parameters..: None
      -------------------------------------------------------------
   */

      PROTECT StringProc
      set procedure to :dUFLP:SetProc.prg additive
      if not SetProc( ":dUFLP:StringEx.cc" )
         set procedure to :dUFLP:stringex.cc additive
         this.StringProc = true
      else
         this.StringProc = false
      endif
      this.String        = new stringex()
      this.String.parent = this
      
   RETURN NULL
   // -- EoM: Init()
   
   FUNCTION Release
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/24/1996
       Notes.......: Used to release this object from memory.
       Written for.: Visual dBASE 5.5
       Rev. History: 11/24/1996 -- Original Code
       Calls.......: none
       Usage.......: convert.Release()
       Example.....: _app.Convert.Release()
                     _app.Convert = NULL  // not required, good 
                                          // idea
       Returns.....: Null
       Parameters..: None
      -------------------------------------------------------------
   */

      this.String.release()
      this.String = null
      if this.StringProc
         close procedure :dUFLP:stringex.cc
      endif
      release object this
   
   RETURN NULL
   // -- EoM: Release()
  
   FUNCTION Arabic( cRoman )
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/09/1996
       Notes.......: This function converts a Roman Numeral to an 
                     arabic one. It parses the roman numeral into 
                     an array, and checks each character ... if 
                     the previous character causes the value to 
                     subtract (for example, IX = 9, not 10) we 
                     subtract that value, and then set the 
                     previous value to 0, otherwise we would get 
                     some odd values in return. So far, it works 
                     fine.
       Written for.: dBASE IV, 1.1
       Rev. History: 07/15/1991 - original function.
                     04/26/1992 - Jay Parsons - shortened (or 
                                  perhaps "tightened the code" 
                                  is better wording).
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Arabic(<cRoman>)
       Example.....: ? _app.Convert.Arabic("XXIV")
       Returns.....: Arabic number (from example, 24)
       Parameters..: cRoman = character string containing roman 
                              numeral to be converted.
      -------------------------------------------------------------
   */
      
      local cRom,cLetrs,nLast,nAt,nVal,cChar,nArabic

      // Minor modification to use built-in string methods --
      // note that string variables can automatically use 
      // these:
      cRom = cRoman.toUpperCase().leftTrim().rightTrim()
                                  // convert to all caps in case
      cLetrs  = "IVXLCDMWY"
      nArabic = 0
      nLast   = 0
      do while len( cRom ) > 0
         cChar = right( cRom, 1 )
         nAt   = at( cChar, cLetrs )
         nVal  = 10 ^ int( nAt/2 ) / iif(nAt/2 = int(nAt/2),2,1)
         do case
            case nAt == 0
                 nArabic := 0
                 exit
            case nAt >= nLast
                 nArabic += nVal
                 nLast   := nAt
            otherwise
                 if nAt/2 == int( nAt / 2 )
                    nArabic := 0
                    exit
                 else
                    nArabic -= nVal
                 endif
         endcase
         cRom := left( cRom, len( cRom ) - 1 )
      enddo
      
   RETURN nArabic
   // -- EoM: Arabic()

   FUNCTION BankRound( nNumber, nPlaces )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Rounds numeric argument to given number of 
                     places, which if positive are decimal places, 
                     otherwise trailing zeroes before the decimal, 
                     in accordance with the special banker's rule 
                     that if the value lost by rounding is exactly 
                     halfway between two possible digits, the 
                     final digit expressed will be even.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992
                     Fall, 1994 - revised for dBASE 5.0 for 
                                  Windows. Jay Parsons
                     11/09/1996 - Modified to work with this class
       Calls.......: None
       Usage.......: BankRound(<nNumber>,<nPlaces>)
       Example.....: ? BankRound( 357.725, 2 )
       Returns.....: Numeric = rounded value ( 357.72 in example )
       Parameters..: nNumber = numeric value to round
                     nPlaces = decimal places, negative being 
                               powers of 10
      -------------------------------------------------------------
   */
      
      local nTemp
      if pCount() < 2
         nPlaces = 1
      endif
      nTemp = nNumber * 10 ^ nPlaces +.5
      if nTemp == int( nTemp ) and nTemp / 2 # int(nTemp / 2)
         nTemp := nTemp - 1
      endif
         
   RETURN int( nTemp ) / 10 ^ nPlaces
   // -- EoM: BankRound()

   FUNCTION Bin2Dec( cStr )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts a string containing a binary value
                     to its numeric (decimal) equivalent.  Any 
                     characters in the string other than "0" or 
                     "1" are ignored.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/25/1992 - original function
                     Fall, 1994 - revised for dBASE 5.0 for 
                                  Windows. Jay Parsons
                     11/09/1996 - Modified to work with this class
       Calls.......: None
       Usage.......: Bin2Dec( <cStr> )
       Example.....: ? Bin2Dec( "1000 0011" )
       Returns.....: Numeric = equivalent ( 131 in example )
       Parameters..: cStr1 = string holding binary value to convert
      -------------------------------------------------------------
   */
      
      local nNext, cChar, nVal
      
      nVal = 0
      for nNext = 1 to len( cStr )
          cChar = substr( cStr, nNext, 1 )
          if cChar $ "01"
            nVal = 2 * nVal + val( cChar )
          endif
      endfor
      
   RETURN nVal
   // -- EoM: Bin2Dec()

   FUNCTION BreakLong( cLong, nPos ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts four characters to unsigned integer, 
                     as is usually appropriate for Windows API 
                     constants. The reason this routine is 
                     required as well as the less-used Long2Signed 
                     is that dBASE interprets a value created by 
                     bitlshift() in which the highest of 32 bits 
                     is a 1 as a signed integer, rather than the 
                     unsigned integer usually intended. The problem
                     does not exist with shorter (WORD) values. 
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this 
                                  class.
       Calls.......: convert.Long2Signed
       Usage.......: convert.BreakLong( <cLong>, <nPos> ) 
       Example.....: ? _app.Convert.BreakLong( cBuffer, 3 ) 
       Returns.....: An integer in the range 0 to 2^32 -1 
                     representing the numeric equivalent of four 
                     characters starting at nPos of cLong 
                     considered as an unsigned long binary number. 
       Parameters..: cLong  = the name of a four-character string 
                              holding the binary value, low-order 
                              byte first 
                     nPos   = the position in cLong of the first 
                              byte 
      -------------------------------------------------------------
   */
  
   RETURN ( mod( CLASS::Long2Signed((cLong),(nPos)), 2^32 ) ) 
   // -- EoM: BreakLong()

   FUNCTION BreakWord( cWord, nPos ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts two characters to unsigned integer.
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 -- Original Release 
                     11/09/1996 -- Modified to work with this class
       Calls.......: None 
       Usage.......: convert.BreakWord( <cWord>, <nPos> ) 
       Example.....: nValue = _app.Convert.BreakWord( fread( ;
                                           Nhandle, 2), 1 )
       Returns.....: An integer in the range 0 to 65535 
                     representing the numeric equivalent of two 
                     characters starting at cWord considered as 
                     an unsigned binary number. 
       Parameters..: cWord  = the name of a two-character string 
                              holding the binary value, low-order 
                              byte first 
                     nPos   = the starting position of the word 
                              within the cWord string 
      -------------------------------------------------------------
   */
 
   RETURN ( asc( substr((cWord),(nPos), 1) ) + ; 
          bitlshift( asc( substr( (cWord), (nPos) + 1, 1 ) ), 8 ) )
   // -- EoM: BreakWord()

   FUNCTION ByteVal(cArg, nPos, nLen)
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan
       Date........: 11/09/1996
       Notes.......: Converts (multi-)byte value stored in a 
                     character string to its numeric value. Used 
                     primarily for interfacing with API.
       Written for.: dBASE 5.0 for Windows
       Rev. History: 07/06/1994 - Original
                     08/15/1994 - Kenneth Chan, changed conversion
                                   algorithm to use built-in BITx()
                                   functions. Set defaults for 
                                   arguments
                     10/06/1994 - Jay Parsons. revised slightly to
                                   leave parameters unchanged
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.ByteVal(<cArg>,<nPos>,<nLen>)
       Example.....: _app.Convert.ByteVal(cRect, 3, 2)
       Returns.....: Numeric
       Parameters..: cArg = character string containing value
                     nPos = position in string where (multi-)byte 
                            value starts
                     nLen = length of value (maximum 4 [32 bits])
      -------------------------------------------------------------
   */

      #ifdef MAX_LEN
         #undef MAX_LEN
      #endif
      #define MAX_LEN 4             // maximum 32 bits
      
      local nBytes, nStart, nRet, nPtr
      
      if pcount() < 3
         nBytes = len(cArg)
         if pcount() < 2
            nStart = 1
         else
            nStart = nPos
         endif
      else
          nBytes = nLen
      endif
      
      nRet = 0
      nBytes = max( 1, min( nLen, MAX_LEN ) )
      /*
         Numbers are stored in little-endian format, with the low
         byte first. The loop reads the characters backward,
         starting at the high byte. Each iteration shifts the
         previous total "up" one byte and "masks-in" the current
         character's value.
      */
      for nPtr = nPos + nBytes - 1 to nPos step - 1
          nRet = bitor( bitlshift( nRet, 8 ), ;
                        asc( substr( cArg, nPtr, 1 )))
      endfor
       
      /*
         The BITx() functions treat their arguments as 32-bit
         numbers, with the high bit as a sign bit. If the result
         of the previous loop is a negative number, that means
         the high bit of a 32-bit number was set, so the number
         must be converted to its unsigned 32-bit value.
      */
      if nRet < 0
         nRet = 256 ^ MAX_LEN + nRet
      endif
         
      #undef MAX_LEN
        
   RETURN nRet
   // -- EoM: ByteVal()

   FUNCTION Cash2Check( nCash,cCurrency,cSmallChange,lPluralCurrency,lpluralSmallChange )
   /*
      -------------------------------------------------------------
       Programmers.: Jay Parsons, Mervyn Bick 
       Date........: 11/09/1996
       Notes.......: Converts a number of dollars and cents to a 
                     string of words appropriate for writing 
                     checks. To correctly evaluate values over 
                     16 decimal places, SET PRECISION TO a value 
                     larger than the default of 16 before calling 
                     this function.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
                     07/06/2010 - Modified by Mervyn Bick to cater for 
                                  foreign currencies
       Calls.......: convert.Num2Words()
                     convert.Thou2Words()
       Called by...: Any
       Usage.......: convert.Cash2Check(<nCash>)
       Example.....: ? _app.Convert.Cash2Check( 348.27 )
       Returns.....: Character string equivalent
       Parameters..: nCash = money value to convert
       Optional Parameters..: cCurrency = main currency unit eg "naira" 
                                          for Nigeria (default is "dollar")
                              cSmallchange = subordinate currency unit eg 
                                             "kobo" for Nigeria (default 
                                             is "cent") 
                              lPluralCurrency = true if currency adds "s" 
                                                for plural (default)
                              lPluralSmallChange = true if subordinate 
                                                   currency adds "s" for 
                                                   plural (default) 
      -------------------------------------------------------------
   */

      local nDollars, nCents, cResult

      if argCount() = 4
         lPluralSmallChange = true
      elseif argcount() = 3
         lPluralSmallChange = true
         lPluralCurrency = true
      elseif argCount() = 2
         lPluralSmallChange = true
         lPluralCurrency = true
         cSmallChange = "cent"
      elseif argCount() = 1
         lPluralSmallChange = true
         lPluralCurrency = true
         cSmallChange = "cent"
         cCurrency = "dollar"
      endif
      nDollars = int( nCash )
      nCents   = 100 * round( nCash - nDollars, 2 )
      cResult  = trim( CLASS::Num2Words( nDollars ) )
      if left( cResult, 1 ) = "C" // deals with oversize number
         RETURN proper( cResult )
      endif
      cResult := proper( cResult + " "+cCurrency + ;
                    iif( nDollars # 1,iif(lPluralCurrency,"s",""), "" ) + " and " )
      if nCents # 0
         RETURN proper( cResult + CLASS::Thou2Words( nCents ) +;
                    " "+cSmallChange + ;
                    iif( nCents # 1,iif(lPluralSmallChange,"s",""), "" ) )
      else
         RETURN proper( cResult + "no "+cSmallChange+iif(lPluralSmallChange,"s",""))
      endif
   // -- EoM: Cash2Check()

   FUNCTION Cel2Fahr( nCelTemp )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Celsius to Fahrenheit 
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Cel2Fahr(<nCelTemp>)
       Example.....: ? _app.Convert.Cel2Fahr(50)
       Returns.....: dBASE number
       Parameters..: nCelTemp -- temperature Celsius
      -------------------------------------------------------------
   */

   RETURN ( 9 * (nCelTemp) / 5 + 32 )
   // -- EoM: Cel2Fahr()

   FUNCTION CM2Inch( nCm )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts centimeters to inches
       Written for.: dBASE IV
       Rev. History: 11/05/1990 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302                                     ,
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.CM2Inch(<nCm>)
       Example.....: ? _app.Convert.CM2Inch(100)
       Returns.....: dBASE number
       Parameters..: nCm -- length in Cm
      -------------------------------------------------------------
   */

   RETURN ((nCm) / 2.54 )
   // --EoM: CM2Inch()

   FUNCTION CuFt2Gal( nCuFt )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Cubic Feet to Gallons (US) 
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.CuFt2Gal(<nCuFt>)
       Example.....: ? _app.Convert.CuFt2Gal(50)
       Returns.....: dBASE number
       Parameters..: nCuFt -- Cubic Feet
      -------------------------------------------------------------
   */

   RETURN ((nCuFt) * 1728 / 231 )
   // -- EoM: CuFt2Gal()

   FUNCTION Dec2Bin( nNum, nPlaces )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts an unsigned integer to a character
                     string giving its ASCII binary representation.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                   Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Dec2Bin(<nNum>,<nPlaces>)
       Example.....: ? _app.Convert.Dec2Bin( 35, 8 )
       Returns.....: Character binary equivalent ( "00100011", in 
                     example)
       Parameters..: nNum    = number to convert
                     nPlaces = number of binary places number is to 
                               occupy
      -------------------------------------------------------------
   */
      
      local cBits, nN
       
      cBits = ""
      nN    = nNum
      do while len( cBits) < nPlaces
         if nN > 0
            cBits := str( mod( nN, 2 ), 1 ) +  cBits
            nN    := int( nN / 2 )
         else
            cBits := "0" + cBits
         endif
      enddo
      
   RETURN cBits
   // -- EoM: Dec2Bin()

   FUNCTION Dec2Mkd( nVar )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a numeric value to eight-character 
                     string. This is roughly equivalent to MKD$()
                     in BASIC. The bytes of the string hold the 
                     value in IEEE long real format ( low-order 
                     byte first.) From high to low, the 64 bits 
                     are:
                           1 bit sign, 1 = negative
                          11 bits exponent base 2 + 1023
                          23 bits mantissa with initial "1." 
                             omitted as understood.
                     dBASE uses this format for floats and dates 
                     internally and in .MEM files; obviously, 
                     the dBASE float() function will make the 
                     same conversion more quickly, but creates 
                     difficulties in accessing the bytes as 
                     converted. This function is designed for 
                     when you must export a number as a float, 
                     such as to a file.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     09/28/1994 - Jay Parsons, revised to write to 
                                  string instead of array
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Bin2Dec()
                     convert.Dec2Bin()
                     convert.Frac2Bin()
       Usage.......: convert.Dec2Mkd( nVar )
       Example.....: ? _app.Convert.Dec2Mkd( -1 )
       Returns.....: character string.
       Parameters..: nVar  = number to convert
      -------------------------------------------------------------
   */
      
      local cStr, cBin, nVal, nExp, nMant, nX
      
      cStr = replicate( chr( 0 ), 8 )
      if nVar # 0
         cBin  = iif( nVar < 0, "1", "0" )
         nVal  = abs( nVar )
         nExp  = int( log( nVar ) / log( 2 ) )
         nMant = nVal / 2 ^ nExp - 1
         cBin  = cBin + CLASS::Dec2Bin( nExp + 1023, 11 ) + ;
                    CLASS::Frac2Bin( nMant, 52 )
         for nX = 1 to 8
             cStr = stuff( cStr, nX, 1, ;
                    chr( CLASS::Bin2Dec( ;
                    substr( cBin, 65 - nX * 8, 8))))
         endfor
      endif
      
   RETURN cStr
   // -- EoM: Dec2Mkd()

   FUNCTION Dec2Mki( nInt )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts an integer in the range -32,768 to 
                     +32,767 to a two-character string equivalent 
                     to the two bytes created by the BASIC MKI$ 
                     function. The same format is used by dBASE 
                     for internal storage of integers within the 
                     range, and by C as a signed int
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     09/28/1994 - Jay Parsons, revised to use 
                                  MakeWord and no array
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class
       Calls.......: convert.MakeWord()
       Usage.......: convert.Dec2Mki( nInt )
       Example.....: ? _app.Convert.Dec2Mki( -1 )
       Returns.....: two-character string containing the value in 
                     binary, low byte first, or empty string for 
                     error
       Parameters..: nInt  = integer to convert
      -------------------------------------------------------------
   */
      
      if nInt < -32768 or nInt > 32767 or nInt # int( nInt )
         RETURN ""
      else
         RETURN CLASS::MakeWord( mod( nInt, 65536 ) ) 
                                      // convert to unsigned
      endif
   // -- EoM: Dec2Mki()

   FUNCTION Dec2Mkl( nInt )
   /*
      --------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts an integer in the range -2^31 to 
                     +2^31 -1 to four-character string.  See 
                     notes to Dec2Mki() Returns null string 
                     if parameter is out of range. This is 
                     equivalent to MKL$() in BASIC.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     09/28/1994 - Jay Parsons, revised to use 
                                  MakeLong and no array
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.MakeLong()
       Usage.......: convert.Dec2Mkl( nInt )
       Example.....: ? _app.Convert.Dec2Mkl( -1 )
       Returns.....: four-character string containing value in 
                     binary, low byte first, or empty string.
       Parameters..: nInt  = integer to convert
      -------------------------------------------------------------
   */
      
      if nInt < -2^31 or nInt > 2^31 - 1 or nInt # int( nInt )
         RETURN ""
      else
         RETURN CLASS::MakeLong( mod( nInt, 2^32 ) ) 
                                         // convert to unsigned
      endif
      
   RETURN cStr
   // -- EoM: Dec2Mkl()

   FUNCTION Dec2Mks( nVar )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a numeric value to four-character 
                     string in IEEE short real format ( low-order 
                     byte first.) This is equivalent to MKS$() 
                     in BASIC. From high to low, the 32 bits are:
                           1 bit sign, 1 = negative
                           8 bits exponent base 2 + 127
                          23 bits mantissa with initial "1." 
                             omitted as understood.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     09/28/1994 - Jay Parsons, revised so no array
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Bin2Dec()
                     convert.Dec2Bin()
                     convert.Frac2Bin()
       Usage.......: convert.Dec2Mks( nVar )
       Example.....: ? _app.Convert.Dec2Mks( -1, )
       Returns.....: four-character string
       Parameters..: nVar  = number to convert
      -------------------------------------------------------------
   */
      
      local cStr, cBin, nVal, nExp, nMant, nX
      
      cStr = replicate( chr(0), 4 )
      if nVar # 0
         cBin = iif( nVar < 0, "1", "0" )
         nVal = abs( nVar )
         nExp = int( log( nVar ) / log( 2 ) )
         nMant = nVal / 2 ^ nExp - 1
         cBin = cBin + CLASS::Dec2Bin( nExp + 127, 8 ) + ;
                       CLASS::Frac2Bin( nMant, 23 )
         for nX = 1 to 4
             cStr = stuff( cStr, nX, 1, ;
                    chr( CLASS::Bin2Dec( ;
                    substr( cBin, 33 - nX * 8, 8))))
         endfor
      endif
      
   RETURN cStr
   // -- EoM: Dec2Mks()

   FUNCTION Dec2MSks( nVar )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a numeric value to four-character 
                     string containing the value in the OBSOLETE 
                     Microsoft short real format. Microsoft now 
                     uses IEEE-standard formats so you should not 
                     need this function unless dealing with older 
                     files using the older format. This is 
                     equivalent to MKS$() in old Microsoft BASIC.
                     From high to low, the 32 bits are:
                           8 bits exponent base 2 + 128
                           1 bit sign, 1 = negative
                          23 bits mantissa with initial ".1" 
                             omitted as understood.
       Written for.: dBASE IV, 1.1
       Rev. History: 12/01/1992 - original function
                     09/28/1994 - Jay Parsons, revised so no array
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Bin2Dec()
                     convert.Dec2Bin()
                     convert.Frac2Bin()
       Usage.......: convert.Dec2MSks( nVar )
       Example.....: ? _app.Convert.Dec2MSks( -1 )
       Returns.....: four-character string holding value, low byte 
                     first
       Parameters..: nVar  = number to convert
      -------------------------------------------------------------
   */      
      local cStr, cBin, nVal, nExp, nMant, nX
      
      cStr = replicate( chr(0), 4 )
      if nVar # 0
         cBin = iif( nVar < 0, "1", "0" )
         nVal = abs( nVar )
         nExp = int( log( nVar ) / log( 2 ) )
         nMant = nVal / 2 ^ nExp - 1
         cBin = CLASS::Dec2Bin( nExp + 129, 8 ) + cBin + ;
                CLASS::Frac2Bin( nMant, 23 )
         for nX = 1 to 4
             cStr = stuff( cStr, nX, 1, ;
                    chr( CLASS::Bin2Dec( ;
                    substr( cBin, 33 - nX * 8, 8))))
         endfor
      endif
      
   RETURN cStr
   // -- EoM: Dec2MSks()

   FUNCTION Dec2Oct( nDec )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts an unsigned integer to its octal 
                     string equivalent
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Dec2Oct(<nDec>)
       Example.....: ? _app.Convert.Dec2Oct( 118 )
       Returns.....: Character = octal string ( "166" in example )
       Parameters..: nDec = number to convert
      -------------------------------------------------------------
   */
      
      local nD, cO
      
      nD = int( nDec )
      cO = ""
      do while nD > 0
         cO = substr( "01234567", mod( nD, 8 ) + 1 , 1 ) + cO
         nD = int( nD / 8 )
      enddo
      
   RETURN iif( "" = cO, "0", cO )
   // -- EoM: Dec2Oct()

   FUNCTION Deg2Num( cDms )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Degrees to number: String in the form 
                     " 40d50'30.2 N"  is converted to a number of 
                     degrees.  If followed by E or S, sign will be 
                     reversed.
      
                     It doesn't matter what characters are used to 
                     separate the degrees, minutes and seconds, but 
                     any of the characters N, E, W and S or their 
                     lowercase equivalents following the last digit 
                     will be understood as specifying a compass 
                     direction.
       
                     If the degrees or minutes are 0, they must 
                     never-the-less be included in the argument.  
                     Seconds may be omitted if 0, as may the 
                     minutes if 0 and seconds are omitted.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1993 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.   
       Calls.......: string.Strip2Val()
                     string.StripVal()
                     string.StrpBrk()
       Usage.......: convert.Deg2Num(<cDms>)
       Example.....: ? _app.Convert.Deg2Num("40d50'30.2 N")
       Returns.....: Numeric (degrees)
       Parameters..: cDms = string holding Degrees Minutes and 
                            Seconds
      -------------------------------------------------------------
   */
   
      local nResult, cStrleft
      if type( "cDms" ) $ "NF"
         RETURN cDms
      endif
      
      cStrLeft = this.String.Strip2Val( cDms )
      nResult  = val( cStrLeft )
      if "" # this.String.Strip2Val( ;
              this.String.StripVal( cStrLeft ) )
         cStrLeft = this.String.Strip2Val( ;
                    this.String.StripVal( cStrLeft ) )
         nResult = nResult + val( cStrLeft ) / 60
         if "" # this.String.Strip2Val( ;
                 this.String.StripVal( cStrLeft ) )
            cStrLeft = this.String.Strip2Val( ;
                       this.String.StripVal( cStrLeft ) )
            nResult = nResult + val( cStrLeft ) / 3600
         endif
      endif
      cStrLeft = upper( ltrim( this.String.StripVal( cStrLeft ) ) )
      if this.String.StrpBrk("NW", cStrLeft ) > 0 or ;
         this.String.StrpBrk( "ES", cStrLeft )=0
         RETURN nResult
      else
         RETURN -nResult
      endif
   // -- EoM: Deg2Num()

   FUNCTION Fahr2Cel( nFahrTemp )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts Fahrenheit to Celsius
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Fahr2Cel(<nFahrTemp>)
       Example.....: ? _app.Convert.Fahr2Cel(50)
       Returns.....: dBASE number
       Parameters..: nFahrTemp -- temperature Fahrenheit
      -------------------------------------------------------------
   */

   RETURN ( 5 * ( (nFahrTemp) - 32 ) / 9 )
   // -- EoM: Fahr2Cel()

   FUNCTION Frac2Bin( nNum, nPlaces )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 03/01/1992
       Notes.......: Converts the fractional part of an unsigned 
                     number to a character string giving its ASCII 
                     binary representation.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Frac2Bin(<nNum>,<nPlaces>)
       Example.....: ? _app.Convert.Frac2Bin( .35, 8 )
       Returns.....: Character binary equivalent
       Parameters..: nNum    = number to convert
                     nPlaces = number of binary places number is to 
                               occupy
      -------------------------------------------------------------
   */
      
      local cBits, nN
      
      cBits = ""
      nN = nNum
      do while len(  cBits ) < nPlaces
         if nN > 0
            nN = 2 * nN
            cBits += str( int( nN ), 1 )
            nN -= int( nN )
         else
            cBits += "0"
         endif
      enddo
      
   RETURN  cBits
   // -- EoM: Frac2Bin()

   FUNCTION Gal2CuFt( nGals )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts Gallons (US) to Cubic Feet
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Gal2CuFt(<nGals>)
       Example.....: ? _app.Convert.Gal2CuFt(50)
       Returns.....: dBASE number
       Parameters..: nGals -- Gallons
      -------------------------------------------------------------
   */
 
   RETURN ((nGals) * 231 / 1728 )
   // -- EoM: Gal2CuFt()

   FUNCTION Gal2Ltr( nGal )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Gallons (US) to Liters
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Gal2Ltr(<nGal>)
       Example.....: ? _app.Convert.Gal2Ltr(50)
       Returns.....: dBASE number
       Parameters..: nGal -- US Gallons
      -------------------------------------------------------------
   */

   RETURN ((nGal) * 3.785411784 )
   // -- EoM: Gal2Ltr()

   FUNCTION Inch2CM( nInches )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts inches to centimeters
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Inch2CM(<nInches>)
       Example.....: ? _app.Convert.Inch2CM(36)
       Returns.....: dBASE number
       Parameters..: nInches -- number of inches
      -------------------------------------------------------------
   */

   RETURN ((nInches) * 2.54 )
   // -- EoM: Inch2CM()

   FUNCTION Kg2Lb( nKG )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts kilograms to pounds
       Written for.: dBASE IV
       Rev. History: 11/05/1990 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Kg2Lb(<nKg>)
       Example.....: ? _app.Convert.Kg2Lb(40)
       Returns.....: dBASE number
       Parameters..: nKg -- mass in kilograms
      -------------------------------------------------------------
   */

   RETURN ((nKG) / 0.45359237 )
   // -- EoM: Kg2Lb()

   FUNCTION Km2Mile( nKM )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 03/01/1992
       Notes.......: converts kilometers to miles
       Written for.: dBASE IV
       Rev. History: Ver. 1.1 eliminates one multiplication
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Km2Mile(<nKm>)
       Example.....: ? _app.Convert.Km2Mile(80)
       Returns.....: dBASE number
       Parameters..: nKm -- length in kilometers
      -------------------------------------------------------------
   */

   RETURN (( nKM ) / 1.609344 )
   // -- EoM: Km2Mile()

   FUNCTION Km2Naut( nKm )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Kilometers to Nautical Miles
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Km2Naut(<nKm>)
       Example.....: ? _app.Convert.Km2Naut(50)
       Returns.....: dBASE number
       Parameters..: nKm -- length in kilometers
      -------------------------------------------------------------
   */

   RETURN ( (nKM ) / 1.852 )
   // -- EoM: Km2Naut()

   FUNCTION Lb2Kg( nLbs )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts pounds to kilograms
       Written for.: dBASE IV
       Rev. History: 11/05/1990 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Lb2Kg(<nLbs>)
       Example.....: ? _app.Convert.Lb2Kg(66)
       Returns.....: dBASE number
       Parameters..: nLbs -- mass in pounds
      -------------------------------------------------------------
   */

   RETURN  ((nLbs) * 0.45359237 )
   // -- EoM: Lb2Kg()

   FUNCTION Long2Signed( cLong, nPos )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts four characters to signed integer. 
                     Most Windows API constants should be converted 
                     to unsigned integers instead, using BreakLong. 
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.BreakWord()
       Usage.......: convert.Long2Signed( <cLong>, <nPos> ) 
       Example.....: ? _app.Convert.Long2Signed( cBuffer, 3 ) 
       Returns.....: An integer in the range -2^31 to 2^31 -1 
                     representing the numeric equivalent of four 
                     characters starting at cLong considered as a 
                     signed long binary number. 
       Parameters..: cLong  = the name of a four-character string
                              holding the binary value sought, 
                              low-order byte first
                     nPos   = the starting position in the string 
      -------------------------------------------------------------
   */
 
   RETURN ( CLASS::BreakWord( (cLong),(nPos) ) + ; 
            bitlshift( CLASS::BreakWord( cLong, nPos) + 2, 16 ) ) 
   // -- EoM: Long2Signed()

   FUNCTION Ltr2Gal( nLiter )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Liters to Gallons (US) 
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Ltr2Gal(<nLiter>)
       Example.....: ? _app.Convert.Ltr2Gal(50)
       Returns.....: dBASE number
       Parameters..: nLiter -- Liters
      -------------------------------------------------------------
   */

   RETURN ((nLiter) / 3.785411784 )
   // -- EoM: Ltr2Gal()

   FUNCTION MakeLong( nLong ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a number in the range 0 to 2^32 -1 to
                     four  characters holding the binary 
                     equivalent. 
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.MakeWord()
       Called by...: Any 
       Usage.......: convert.MakeLong( <nLong> ) 
       Example.....: fwrite( nHandle, ;
                             _app.Convert.MakeLong( nValue ), 4) 
       Returns.....: A four-character string holding the value in 
                     binary, low-order byte first. 
       Parameters..: nLong  = the number, from 0 to 2^32 -1, to 
                              convert 
      -------------------------------------------------------------
   */

   RETURN (CLASS::MakeWord( bitand( nLong, 65535) ) + ; 
           CLASS::MakeWord( bitrshift( nLong, 16) ) ) 
   // -- EoM: MakeLong()
 
   FUNCTION MakeWord( nWord ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a number holding a value from 0 to 
                     65535 into two characters holding the binary 
                     equivalent.
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this class.
       Calls.......: None 
       Usage.......: convert.MakeWord( <nWord> ) 
       Example.....: ? fwrite(nHandle,;
                                 _app.Convert.MakeWord( 35277), 2 )
       Returns.....: A two-character string holding the value in 
                     binary, low-order byte first. 
       Parameters..: nWord = the number, from 0 to 65535, to 
                             convert 
      -------------------------------------------------------------
   */

   RETURN ( chr( bitand( nWord, 255 ) ) + ;
            chr( bitrshift( nWord, 8 ) ) ) 
   // -- EoM: MakeWord()

   FUNCTION Mile2KM( nMiles )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts Miles to Kilometers
       Written for.: dBASE IV
       Rev. History: Ver. 1.1 eliminates one multiplication
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Mile2KM(<nMiles>)
       Example.....: ? _app.Convert.Mile2KM(50)
       Returns.....: dBASE number
       Parameters..: nMiles -- length in miles
      -------------------------------------------------------------
   */

   RETURN ((nMiles) * 1.609344 )
   // --EoM: Mile2KM()

   FUNCTION Mkd2Dec( cString )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts an eight-character string storing an 
                     IEEE long real value ( as saved by the BASIC 
                     MKD$ function, e. g. ) to its numeric 
                     (decimal) equivalent.  The string must store 
                     the value low-order byte first. 
                     From high to low, the 64 bits are:
                           1 bit sign, 1 = negative
                          11 bits exponent base 2 + 1023
                          52 bits mantissa with initial "1." 
                              omitted as understood.
                     This is the equivalent of CVD() in BASIC.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     04/12/1993 - Jay Parsons, changed to work 
                                  around dBASE IV 2.0 mod() bug
                     09/29/1994 - Jay Parsons, changed to use 
                                  string and convert it to binary 
                                  in reverse
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Bin2Dec()
       Usage.......: convert.Mkd2Dec( <cString> )
       Example.....: ? _app.Convert.Mkd2Dec( chr( 0 ) + chr( 0 ) +; 
                                           chr( 0 ) + chr( 0 ) + ;
                                           chr( 0 ) + chr( 0 ) + ;
                                           chr( 248 ) + chr( 3 )
       Returns.....: Numeric = equivalent ( 1 in example )
       Parameters..: cString = string holding value to convert
      -------------------------------------------------------------
   */
      
      local nX, nY, cBin, nSign, nExp, cMant, nVal, nZ
      
      nZ = 0
      cBin = ""
      for nX = 1 to 8
          nVal = asc( substr( cString, nX, 1 ) )
          nZ += nVal
          for nY = 0 to 7
              cBin = str( mod( nVal, 2 ), 1 ) + cBin
              nVal = bitrshift(  nVal, 1 )
          endfor
      endfor
      if nZ = 0
         nVal = 0
      else
         nSign = iif( left( cBin, 1 ) == "1", -1, 1 )
         nExp  = class::Bin2Dec( substr( cBin, 2, 11) ) - 1023
         cMant = "1" + right( cBin, 52 )
         nVal  = class::Bin2Dec( cMant ) * ;
                 2 ^ ( nExp - 52 ) * nSign
      endif
      
   RETURN nVal
   // -- EoM: Mkd2Dec()

   FUNCTION Mki2Dec( cString )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/06/1996
       Notes.......: Converts two-char string storing a signed 
                     integer ( as saved by the BASIC MKI$ 
                     function, e. g. ) to its numeric (decimal) 
                     equivalent.  The format accommodates values 
                     from 8000 ( -32,768 ) to 7FFF ( +32,767 ); 
                     the low-order byte is stored first.
                           This is the equivalent of CVI() in 
                           BASIC.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/25/1992 - original function
                     09/29/1994 - Jay Parsons, revised to use 
                                  string
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.BreakWord()
       Usage.......: convert.Mki2Dec( <cString> )
       Example.....: ? _app.Convert.Mki2Dec( chr( 255 ) + ;
                                             chr( 255 ) )
       Returns.....: Numeric = equivalent ( -1 in example )
       Parameters..: cString = string holding value to convert
      -------------------------------------------------------------
   */
      
      local nRet
      nRet = CLASS::BreakWord( cString, 1 )
      if nRet > 32767
         nRet = nRet - 32768
      endif
      
   RETURN nRet
   // -- EoM: Mki2Dec()

   FUNCTION Mkl2Dec( cString )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a four-character string storing a 
                     signed long integer ( as saved by the BASIC 
                     MKL$ function, e. g. ) to its numeric 
                     (decimal) equivalent.  The string must
                     have the low-order byte stored first.
                           This is the equivalent of CVL() in 
                           BASIC.
       Written for.: dBASE IV, 1.1
       Rev. History: 11/26/1992 - original function
                     09/29/1994 - Jay Parsons, revised to use 
                                  string
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Long2Signed
       Usage.......: convert.Mkl2Dec( <cString> )
       Example.....: ? _app.Convert.Mkl2Dec( chr( 255 ) + ;
                                chr( 255 ) + ;
                                chr(255 ) + chr( 255) )
       Returns.....: Numeric = equivalent ( -1 in example )
       Parameters..: cString = chars holding value to convert
      -------------------------------------------------------------
   */
   
   RETURN CLASS::Long2Signed( cString, 1 )
   // -- EoF: Mkl2Dec()

   FUNCTION MSks2Dec( cString )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts four-character string storing an old-
                     style Microsoft short real value ( as saved by the 
                     old BASIC MKS$ function, e. g. ) to its 
                     numeric (decimal) equivalent. The four 
                     characters store the value low-order byte 
                     first, From high to low, the 32 bits are:
                           8 bits exponent base 2 + 128
                           1 bit sign, 1 = negative
                          23 bits mantissa with initial ".1" 
                               omitted as understood.
                     This is the equivalent of CVS() in old 
                     Microsoft BASIC
       Written for.: dBASE IV, 1.1
       Rev. History: 11/28/1992 - original function
                     04/12/1993 - changed to work around dBASE IV 
                                  2.0 mod() bug, Jay Parsons
                     09/29/1994 - Jay Parsons, changed to use 
                                  string and convert it to binary 
                                  in reverse
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons, 72662,1302
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Bin2Dec()
       Usage.......: convert.MSks2Dec( <cString> )
       Example.....: ? _app.Convert.MSks2Dec( chr( 0 )+chr( 0 )+;
                                            chr( 128 )+chr( 63 ) )
       Returns.....: Numeric = equivalent ( 1 in example )
       Parameters..: cString = string holding value to convert
      -------------------------------------------------------------
   */
      
      local nX, nY, nZ, cBin, nSign, nExp, cMant, nVal
      
      nZ = 0
      cBin = ""
      for nX = 1 to 4
          nVal = asc( substr( cString, nX, 1 ) )
          nZ += nVal
          for nY = 0 to 7
              cBin = str( mod( nVal, 2 ), 1 ) + cBin
              nVal = bitrshift(  nVal, 1 )
          endfor
      endfor
      if nZ = 0
         nVal = 0
      else
         nSign = iif( substr( cBin, 9, 1 ) = "1", -1, 1 )
         nExp  = CLASS::Bin2Dec( left( cBin, 8 ) ) - 128
         cMant = "1" + right( cBin, 23 )
         nVal = CLASS::Bin2Dec( cMant ) * ;
                     2 ^ ( nExp - 24 ) * nSign
      endif
      
   RETURN nVal
   // -- EoM: MSks2Dec()

   FUNCTION Naut2KM( nNautMiles )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: converts Nautical Miles to Kilometers
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.   
       Calls.......: None
       Usage.......: convert.Naut2KM(<nNautMiles>)
       Example.....: ? _app.Convert.Naut2KM(50)
       Returns.....: dBASE number
       Parameters..: nNautMiles -- length in miles
      -------------------------------------------------------------
   */

   RETURN ( (nNautMiles) * 1.852 )
   // -- EoM: Naut2KM()

   FUNCTION Naut2Stat( nNautMiles )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts Nautical Miles to Statute Miles
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Naut2Stat(<nNautMiles>)
       Example.....: ? _app.Convert.Naut2Stat(50)
       Returns.....: dBASE number
       Parameters..: nNautMiles -- length in miles
      -------------------------------------------------------------
   */

   RETURN ( (nNautMiles) * 1.150779448 )
   // -- EoM: Naut2Stat()

   FUNCTION Num2Real( nNum, nBytes )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a number to the ASCII representation 
                     of its storage in IEEE 4 or 8-byte real 
                     format, with least significant byte (lowest 
                     in memory) first. Provided for checking the 
                     values in .MEM files, or in memory float-
                     type variables if peeking.
       Written for.: dBASE IV Version 1.5
                     ( may be adapted to earlier versions by 
                       requiring fixed number of parameters.)
       Rev. History: 03/01/1992 - original function
                     11/26/1992 - Jay Parsons, revised to call 
                                  Dec2Mkd(), etc.
                     09/28/1994 - Jay Parsons, revised to call 
                                  itoh() and to eliminate array 
                                  use
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Dec2Mkd()
                     convert.Dec2Mks()
       Usage.......: convert.Num2Real(<nNum> [,<nBytes>] )
       Example.....: ? _app.Convert.Num2Real( 10E100, 8 )
       Returns.....: Character string equivalent ( of a blank date,
                     in example )
       Parameters..: nNum = number to convert
                     nBytes = number of bytes in conversion.  
                              Optional,will be considered 8 
                              ( long real ) unless 4 is specified.
      -------------------------------------------------------------
   */      

      local cStr, nB, nX, aMK
      
      nB = iif( type( "nBytes" ) = "N" .AND. nBytes = 4, 4, 8 )
      declare aMK[ nB ]
      cStr = ""
      if "" # iif( nB = 8, CLASS::Dec2Mkd( nNum, "MK" ), ;
                   CLASS::Dec2Mks( nNum, "MK" ) )
         for nX = 1 to nB
             cNext = itoh( asc( aMK[ nX ] ) )
             cStr += right( "0" + ;
                    itoh( asc( aMK[ nX ] ) ), 2 ) + " "
         endfor
      endif
      
   RETURN trim( cStr )
   // -- EoM: Num2Real()

   FUNCTION Num2Str( nNumber )
   /*
      -------------------------------------------------------------
       Programmer..: Angus Scott-Fleming
       Date........: 11/09/1996
       Notes.......: Converts a number to a string like str(), 
                     storing all decimal places. Does not require 
                     knowing the number of decimal places first. 
                     You may have to change the PRECISION setting 
                     before calling this function to avoid loss 
                     of digits.
       Written for.: dBASE IV, 1.1
       Rev. History: 06/09/1992 - Angus took Jay's routine and 
                                  overhauled it.
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Called by...: Any
       Usage.......: convert.Num2Str(<nNumber>)
       Example.....: ? _app.Convert.Num2Str( 415.25000000000001 )
       Returns.....: Character = representation of number 
                                 ( "415.2500000000001" in example )
       Parameters..: nNumber = number to represent
      -------------------------------------------------------------
   */
      
      local nInteger, nFraction, cFracstr, nDec
       
      nInteger  = int( nNumber )
      nFraction = abs( nNumber - nInteger )
      if nFraction == 0
         nFracStr = ""
      else
         // note that the maximum # of decimals is 18
         nFracStr = ltrim(str(nFraction,19,18))
         do while right(nFracStr,1) = "0"
            nFracStr := left(nFracStr,len(nFracStr)-1)
         enddo
      endif
      
   RETURN ltrim( str( nInteger ) ) + nFracStr
   // -- EoM: Num2Str()

   FUNCTION Num2Words( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts an integer to a string of words. 
                     Limited, due to 254-character limit of 
                     dBASE IV strings, to numbers less than 
                     10 ^ 15
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.Thou2Words()
       Usage.......: convert.Num2Words(<nNum>)
       Example.....: ? _app.Convert.Num2Words( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */
            
      local nNumleft, nScale, nGroup, cResult
      
      nNumLeft = int( nNum )
      do case
         case abs( nNumLeft ) >= 10 ^ 15
              RETURN "Cannot convert a number in or above "+;
                     "the quadrillions."
         case nNumLeft == 0
              RETURN "zero"
         case nNumLeft < 0
              cResult = "minus "
              nNumLeft = -nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale   = int( log10( nNumLeft ) / 3 )
         nGroup   = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         cResult += CLASS::Thou2Words( nGroup )
         if nScale > 0
            cResult += " " ;
              + trim( substr( "thousandmillion billion trillion",;
                nScale * 8 - 7, 8 ) )
            if nNumLeft > 0
               cResult += ", "
            endif
         endif
      enddo           
         
   RETURN cResult
   // -- EoM: Num2Words()

   FUNCTION NumVal( cArg )
   /*
      -------------------------------------------------------------
       Programmer..: Kelvin Smith
       Date........: 11/09/1996
       Notes.......: Equivalent to VAL(), but return value only if 
                     string is all numeric.  The string is 
                     considered numeric if it contains digits, a 
                     maximum of one dot, a minus sign only at the 
                     beginning of the string, and/or spaces at the 
                     beginning and/or end of the string, but not 
                     between characters. Note that certain 
                     character strings which may be intended as 
                     numbers will be rejected by this function. In 
                     particular, scientific notation (such as "
                     1E3" for 1000) is considered non-numeric.  
                     This is intentional for my purposes 
                     (emulating the results of a non-dBASE 
                     environment). Obviously, this function does 
                     not distinguish (in its results) between a 
                     non-numeric string and a string with the 
                     value 0.  This function could be altered to 
                     return lValid instead, which is .T. if the 
                     string is numeric and .F. if the string is 
                     non-numeric.
       Written for.: dBASE 5.0
       Rev. History: 09/29/1994 - Original Release
                     11/09/1996 - Modified to work with this class.
       Calls.......: string.AllTrim()
       Usage.......: convert.NumVal(<cval>)
       Example.....: _app.Convert.NumVal("73E5") returns 0
                   : _app.Convert.NumVal("  -34.443  ") 
                                                 returns -34.443
       Returns.....: 0 if non-numeric characters found, 
                                                 else Val(cArg)
       Parameters..: cArg = Character string with possible numeric 
                            value
      -------------------------------------------------------------
   */

      private cChar, nChar, lDot, lValid, nRet, nStart
      
      lDot   = false
      lValid = true                      // Until proven otherwise
      cArg = this.String.AllTrim(cArg)
      if asc(cArg) = 45                  // If leading minus,
         nStart = 2                      // skip it for testing
      else
         nStart = 1
      endif
      for nChar = nStart to len(cArg)
          cChar = substr(cArg, nChar, 1)
          do case
             case cChar == "."
                  if lDot                 // Already have one
                     lValid = false
                     exit
                  endif
                  lDot = true
             case not cChar $ "0123456789"
                  lValid = false
                  exit
          endcase
      next
      if lValid
         nRet = val(cArg)
      else
         nRet = 0
      endif
      
   RETURN nRet
   // -- EoM: NumVal()

   FUNCTION Oct2Dec( xOct )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Converts an unsigned number in octal, or its 
                     string representation, to a numeric (decimal) 
                     value
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/92 - original function.
                     12/01/92 - modified to eliminate usually-
                                harmless "substring out of range" 
                                error, Jay Parsons
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Oct2Dect(<xOct>)
       Example.....: ? _app.Convert.Oct2Dec( 166 )
       Returns.....: Numeric = decimal equivalent ( 118 in example)
       Parameters..: xOct = octal character string or number to 
                            convert
      -------------------------------------------------------------
   */
      
      local nD, cO, cVal
      
      if type( "xOct" ) $ "NF"
         cO = str( xOct )
      else
         cO = xOct
      endif
      cO = upper( trim( ltrim( cO ) ) ) + "!"
      nD = 0
      do while len( cO ) > 1
         cVal = left( cO, 1 )
         if cVal # " "
            nD = nD * 8 + at( cVal, "1234567" )
         endif
         cO = substr( cO, 2 )
      enddo
      
   RETURN nD
   // -- EoM: Oct2Dec()

   FUNCTION Ord( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts an integer to ordinal representation 
                     by adding "st", "nd", "rd" or "th" after its 
                     digit(s)
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Called by...: Any
       Usage.......: convert.Ord(<nNum>)
       Example.....: ? _app.Convert.Ord( 11 )
       Returns.....: Character ordinal string equivalent ( "11th" 
                     in example )
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */
         
      local nD
      
      nD = mod( nNum, 100 ) - 1
                   // the -1 just happens to simplify what follows
      
   RETURN str( nNum ) + iif( mod( nD, 10 ) > 2 or ;
          abs( nD - 11 ) < 2, "th", ;
          substr( "stndrd", mod( nD, 10 ) *;
          2 + 1, 2 ) )
   // -- EoM: Ord()

   FUNCTION Ordinal( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Returns ordinal string for a positive integer 
                     < 100. For higher numbers, use Num2Words on 
                     int( n/100 ),  then use this on mod( n, 100 ) 
                     or if mod( n, 100 ) = 0, add "th" ).
       Written for.: dBASE IV, 1.1
       Rev. History: 11/19/1992 - original function
                     12/03/1992 - Jay Parsons, changed notes and 
                                  variable names, replaced five 
                                  lines with an "iif" line
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Ordinal( <nNum> )
       Example.....: ? _app.Convert.Ordinal( 31 ) 
                                 && returns "thirty-first"
       Returns.....: String giving ordinal value ( position ) of 
                     number, or null
       Parameters..: nNum = integer > 0 and < 100
      -------------------------------------------------------------
   */
      
      local cUnits, cTeens, cDecades, nRest, cOrd
      
      //- 6     12345612345612345612345612345612345612345612345
      cUnits = "     four  fif   six   seven eigh  nin   ten   "+;
               "eleventwelf "
      //- 5       1234512345123451234512345123451234512345
      cTeens =   "    thir four fif  six  seveneigh nine  "
      cDecades = "    twen thir for  fif  six  seveneigh nine"
      
      nRest = nNum
      cOrd = ""
      if nRest # int( nRest ) OR nRest < 1 OR nRest > 99
         nRest = 0
      endif
      
      if nRest > 19
         cOrd = trim( substr( cDecades, 5 * ;
                    ( int( nRest / 10 ) - 1 ), 5 ) ) + "t"
         nRest = mod( nRest, 10 )
         cOrd += iif( nRest = 0, "ieth", "y-" )
      endif
      
      do case
         case nRest > 12
              cOrd += trim( substr( cTeens, 5 * ;
                      ( nRest - 12 ), 5 ) ) + "teenth"
         case nRest > 3
              cOrd += trim( substr( cUnits, ;
                      6 * ( nRest - 3 ), 6 ) ) + "th"
         case nRest > 0
              cOrd += trim( substr( "     first secondthird ", ;
                      6 * nRest, 6 ) )
      endcase
      
   RETURN cOrd
   // -- EoM: Ordinal()

   FUNCTION Roman( nArabic )
   /*
      -------------------------------------------------------------
       Programmer..: Nick Carlin
       Date........: 11/09/1996
       Notes.......: A function designed to return a Roman Numeral 
                     based on an Arabic Numeral input ...
       Written for.: dBASE III+
       Rev. History: 04/13/1988 - original function.
                     07/25/1991 - Ken Mayer - 1) modified for dBASE
                                  IV, 1.1, 2) updated to a 
                                  function, and 3) the procedure 
                                  GetRoman was done away with 
                                  (combined into the function).
                     04/26/1992 - Jay Parsons - shortened 
                                  (seriously)
                     08/27/1993 - Jay Parsons - dBASE IV 2.0 bug 
                                  worked around
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: Roman(<nArabic>)
       Example.....: ? Roman(32)
       Returns.....: Roman Numeral (character string) equivalent of
                     Arabic numeral passed to it. In example:  
                     XXXII
       Parameters..: nArabic = Arabic number to be converted to 
                               Roman
      -------------------------------------------------------------
   */
      
      local cLetrs,cRoman,nCount,nLeft,nMod,nNines,cAdd
       
      cLetrs ="IVXLCDMWY"     // Roman digits
      cRoman = ""             // this will be the returned value
      nCount = 0              // init counter
      nLeft = fixed( nArabic )
      if nLeft < 0 or nLeft # int( nLeft )
         RETURN cRoman
      endif
      do while nCount < 4 and nLeft > 0    // loop four times, once
                                           // each for 1s, 10s, 
                                           // 100s, 1000s
         nMod = mod( nLeft, 10 )
         nLeft = int( nLeft / 10 )
         cGroup = substr( cLetrs, 2 * nCount + 1, 3 )
         cAdd = ""
         do case
            case nMod = 9
                 cAdd = left( cGroup, 1 ) + right( cGroup, 1 )
            case nMod = 4
                 cAdd = left( cGroup, 2 )
            otherwise
                 if nMod > 4                    // 5 - 8
                    cAdd = substr( cGroup, 2, 1 )
                    nMod = nMod - 5
                 endif
                 if nMod > 0                    // 1 - 3 and 6 - 8
                    cAdd = cAdd + replicate(left( cGroup, 1 ), nMod)
                 endif
         endcase
         cRoman = cAdd + cRoman
         nCount++
      enddo  // while nCounter < 4
           
   RETURN cRoman
   // -- EoM: Roman()

   function RGBToHex( cRGB )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: Unknown, added to library 10/4/98
       Notes.......: Converts RGB returned by the GetColor()
                     function values to hexidecial
                     needed for standard RGB strings
                     used in HTML and VdBASE
       Written for.: Visual dBASE 
       Rev. History: 10/04/1998 -- First added to dUFLP lib.
       Calls.......: None
       Usage.......: RGBToHex( getColor() )
       Example.....: ? RGBToHex( "0,0,128" )
       Returns.....: Hex string
       Parameters..: cRGB -- value to convert
      -------------------------------------------------------------
   */

      local nComma, nVal, cStr, cRet
      cStr = cRGB

      // Red:
      nComma = at( ",", cStr )
      nVal   = val( left( cStr, nComma - 1 ) )
      cStr  := substr( cStr, nComma + 1 )
      cRet   = itoh( nVal, 2 ) // built-in integer-to-hex

      // Green:
      nComma := at( ",", cStr )
      nVal   := val( left( cStr, nComma - 1 ) )
      cStr   := substr( cStr, nComma + 1 )
      cRet   = itoh( nVal, 2 ) + cRet

      // Blue:
   RETURN "0x"+itoh( val( cStr ), 2 ) + cRet
   // -- EoM: RGBtoHex()

   FUNCTION Stat2Naut( nStatMiles )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: converts Statute Miles to Nautical Miles
       Written for.: dBASE IV
       Rev. History: 03/01/1992 - Original
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: Stat2Naut(<nStatMiles>)
       Example.....: ? Stat2Naut(50)
       Returns.....: dBASE number
       Parameters..: nStatMiles -- length in miles
      -------------------------------------------------------------
   */

   RETURN ( (nStatMiles) / 1.150779448 )
   // -- EoM: Stat2Naut()

   FUNCTION Str2Float
   /*
      -------------------------------------------------------------
      Programmer..: Kelvin Smith 
      Date........: 09/26/1999
      Notes.......: Convert a string which contains a 4- or 8-byte IEEE
                    floating point number to its actual numeric value.
      Written for.: dBASE 5.0 for DOS
      Rev. History: 09/26/1999 1.0 - Original version
      Calls.......: None
      Called by...: Any
      Usage.......: Str2Float(string)
      Example.....: ? Str2Float(cFloat)
      Returns.....: Numeric floating point value
      Parameters..: cFloat - 4- or 8-byte string

      Standard IEEE format for 4-byte (single precision) floating point:
      1 sign bit, 8 bits 2's exponent (offset by 127), 23 bits mantissa 
      (with implied 1 in first bit).  Bytes stored in reverse order.
      Standard IEEE format for 8-byte (double precision) floating point:
      1 sign bit, 11 bits 2's exponent (offset by 1023), 52 bits mantissa.
      Visual dBase includes bit manipulation functions that make this
      easier and more efficient.  Such code is in comments below.
      -------------------------------------------------------------
   */

      PARAMETERS cFloat
      PRIVATE i, nExp, nMantissa, nFloat, lZero
      DECLARE cFlByte[8]
      lZero = true

      DO CASE
         CASE LEN(cFloat) = 4
             * Separate out the bytes
             FOR i = 1 to 4
                 cFlByte[i] = Substr(cFloat, i, 1)
                 lZero = lZero and (Asc(cFlByte[i]) = 0)
             NEXT
             IF lZero
                 nFloat = 0
             ELSE
                 nExp = BitAnd(Asc(cFlByte[4]), 127) * 2;
                        + iif(Asc(cFlByte[3]) > 128, 1, 0) - 127
                 nMantissa = 0
                 for i = 1 to 3
                     nMantissa = nMantissa / 256 + Asc(cFlByte[i])
                 next 
                 nMantissa = nMantissa - iif(nMantissa >= 128, 128, 0)
                 nMantissa = nMantissa / 128 + 1
                 * Assemble the number; mantissa times 2's exponent; 
                   check sign bit
                 nFloat = nMantissa * 2 ^ nExp * ;
                          iif(Asc(cFlByte[4]) >= 128, -1, 1)
             ENDIF
         CASE LEN(cFloat) = 8
             * Separate out the bytes
             FOR i = 1 to 8
                 cFlByte[i] = Substr(cFloat, i, 1)
                 lZero = lZero .AND. (Asc(cFlByte[i]) = 0)
             NEXT
             IF lZero                      // If all bits are off,
                 nFloat = 0                //    value is 0
             ELSE
                 nExp = BitAnd(Asc(cFlByte[8]), 127) * 2 ^ 4;
                        + BitAnd(Asc(cFlByte[7]), 240) / 16 - 1023
                 nMantissa = 0
                 for i = 1 to 6
                     nMantissa = nMantissa / 256 + Asc(cFlByte[i])
                 next 
                 nMantissa = nMantissa / 256 + Mod(Asc(cFlByte[7]), 16)
                 nMantissa = nMantissa / 16 + 1
                 * Assemble the number; mantissa times 2's exponent; 
                   check sign bit 
                   // changed the bit to 8, based on a bug
                   // report/fix by Marc VdB ....
                 nFloat = nMantissa * 2 ^ nExp * ;
                          iif(Asc(cFlByte[8]) >= 128, -1, 1)
             ENDIF
         OTHERWISE
             ? "Illegal string length ", LEN(cFloat),;
               " for StrFloat() - must be 4 or 8 characters."
             nFloat = 0.
      ENDCASE

   Return nFloat
   // -- EoM: StrFloat() 

   FUNCTION StuffLong( cString, nPos, nLong ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Stuffs the binary equivalent of a number from 
                     0 to 2^32 -1 into four characters of a string 
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.MakeLong()
       Usage.......: convert.StuffLong( <cString>, <nPos>, ;
                                                    <nLong> ) 
       Example.....: cStr = _app.Convert.StuffLong( cStr, 8, ;
                                                        40135277 ) 
       Returns.....: The string holding the value in binary, 
                     low-order byte first. 
       Parameters..: cString = the string to stuff
                     nPos    = the starting position in cString
                     nLong   = the number to be stuffed
      -------------------------------------------------------------
   */

   RETURN (stuff( (cString), (nPos), 4, CLASS::MakeLong(nLong) )) 
   // -- EoM: StuffLong()

   FUNCTION StuffWord( cString, nPos, nWord ) 
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons 
       Date........: 11/09/1996
       Notes.......: Stuffs the binary equivalent of a number from 
                     0 to 65535 into two characters of a string 
       Written for.: dBASE 5.0 for Windows 
       Rev. History: 10/01/1994 - Original Release 
                     11/09/1996 - Modified to work with this class.
       Calls.......: convert.MakeWord()
       Usage.......: convert.StuffWord( <cString>, <nPos>, ;
                                                          <nWord> ) 
       Example.....: cStr = _app.Convert.StuffWord( cStr, 5,;
                                                          35277 ) 
       Returns.....: The string holding the value in binary, 
                     low-order byte first. 
       Parameters..: cString = the string to stuff
                     nPos    = the starting position in cString
                     nWord   = the number to be stuffed
      -------------------------------------------------------------
   */

   RETURN (stuff( (cString), (nPos), 2, CLASS::MakeWord(nWord) )) 
   // -- EoM: StuffWord 

   FUNCTION SU( cField, nConvert )
   /*
      -------------------------------------------------------------
       Programmer..: Oktay Amiry
       Date........: 11/09/1996
       Notes.......: Standard Units is a unit of measurement often 
                     referred to in various statistical 
                     calculations. Suffice it to note that SU is 
                     an intrinsic way of looking at data, 
                     indicating whether a value is above or below 
                     the average. A positive SU indicates the 
                     value was above average, while a negative 
                     SU indicates a below average value. Printed 
                     in TechNotes, December 1992.
       Written for.: dBASE IV, 1.5
       Rev. History: 12/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: SU(<cField>,<nConvert>)
       Example.....: ?SU("SCORE",75)
       Returns.....: numeric
       Parameters..: cField   = a numeric field
                     nConvert = number to be converted
      -------------------------------------------------------------
   */
      
      local nAverage,nStandard
      calculate avg( &nField. ) to nAverage, ;
                std( &nField. ) to nStandard
      
   RETURN iif(nStandard # 0,(nNum-nAverage)/nStandard,0)
   // -- EoF: SU()

   FUNCTION Thou2Words( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jay Parsons
       Date........: 11/09/1996
       Notes.......: Converts a positive integer less than 1000 to 
                     a string of characters.
       Written for.: dBASE IV, 1.1
       Rev. History: 03/01/1992 - Original Release
                     Fall, 1994 - revised for dBASE 5.0 for Windows
                                  Jay Parsons
                     11/09/1996 - Modified to work with this class.
       Calls.......: None
       Usage.......: convert.Thou2Words(<nNum>)
       Example.....: ? _app.Convert.Thou2Words( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */
      
      local cUnits, cTens, nN, cResult
      cUnits = "one      two      " ;
             + "three    four     " ;
             + "five     six      " ;
             + "seven    eight    " ;
             + "nine     ten      " ;
             + "eleven   twelve   " ;
             + "thirteen fourteen " ;
             + "fifteen  sixteen  " ;
             + "seventeeneighteen " ;
             + "nineteen "
      cTens = "twen thir for  fif  six  seveneigh nine "
      nN = int( nNum )
      if nN = 0
         RETURN "zero"
      endif
      nResult = ""
      if nNum > 99
         nResult = trim( substr(cUnits, int(nNum / 100 ) * ;
                            9 - 8, 9 ) ) + " hundred"
         nN = mod( nN, 100 )
         if nN = 0
            RETURN nResult
         else
            nResult += " "
         endif
      endif
      if nN > 19
         nResult += trim( substr( cTens, ;
                     int( nN / 10 ) * 5 - 9, 5 ) ) + "ty"
         nN = mod( nN, 10 )
         if nN = 0
            RETURN nResult
         else
            nResult += "-"
         endif
      endif
      
   RETURN nResult + trim( substr( cUnits, nN * 9 - 8, 9 ) )
   // -- EoM: Thou2Words()

FUNCTION Num2FrenchWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jan-Pierre Martel
       Date........: 04/11/2005
       Notes.......: Converts integer numbers to French word 
                     strings. Limited to numbers from -(10^18) 
                     to 10^19. 
         				Based on Num2Words by Jay Parsons.
       Written for.: Visual dBASE PLUS
       Rev. History: 06/22/1998 - Original Release
                     06/25/1998 - Returns trimmed string 
                     February, 2021 -- Updated by Gaetano De Luisi
                        to correct some changes in the language
                        (1990 spelling reform), and some incorrect
                        translations.
       Calls.......: convert.Thou2FrenchWords()
       Usage.......: convert.Num2FrenchWords(<nNum>)
       Example.....: ? _app.Convert.Num2FrenchWords( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local nNumleft, nScale, nGroup, cResult

      nNumLeft = int( nNum )
      do case
         case abs( nNumLeft ) >= 10 ^ 18
              return("Imposible convertir un nombre plus grand "+;
                     "que 999,999,999,999,999,999")
         case nNumLeft == 0
              return("zro")
         case nNumLeft < 0
              cResult = "moins "
              nNumLeft = - nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale   = int( log10( nNumLeft ) / 3 )
         nGroup   = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         if nGroup = 1 and nScale = 1
            cResult += ""
         else
         cResult += CLASS::Thou2FrenchWords( nGroup )
         endif
         if nNumLeft > 0 and cResult.right(2) == "un"
            cResult =  cResult.substring( 0, cResult.length )
         endif
         if nScale = 1 and nGroup=1
				cResult +=  trim( substr( "mille       million     milliard    billion     billiard    ",;
                nScale * 12 - 11, 12 ))+IIF(nNumLeft>0,"-","")
			elseif nScale > 0
				// "mille is never preceded by "un" when at start of number
				// we say "thousand" and not "one thousand" in French
            cResult +="-"+ trim( substr( "mille       million     milliard    billion     billiard    ",;
                nScale * 12 - 11, 12 ) )
            if nGroup # 1 or nScale < 1					
               if "mille " $cResult
						//"mille" is invariable as it is already a plural - don't add an "s" for plural						
					else
						cResult+="s"
					endif
            endif
            if nNumLeft > 0
               cResult += "-"
            endif			
         endif
      enddo

   RETURN cResult
   // -- EoM: Num2FrenchWords()

   FUNCTION Num2GermanWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Erik Moeller
       Date........: 06/22/1998
       Notes.......: Converts integer numbers to German word 
                     strings. Limited to numbers from -(10^19) 
                     to 10^19. 
         				Based on Num2Words by Jay Parsons.
       Written for.: Visual dBASE 7
       Rev. History: 06/22/1998 - Original Release
                     06/25/1998 - Returns trimmed string 
       Calls.......: convert.Thou2GermanWords()
       Usage.......: convert.Num2GermanWords(<nNum>)
       Example.....: ? _app.Convert.Num2GermanWords( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local nNumleft, nScale, nGroup, cResult
	
      if nnum >= 10^19 OR nnum <= -(10^19)
         RETURN "Maximalwert: 10^19. Minimalwert: -(10^19)." 
      endif
      nNumLeft = int( nNum ) // must be done here, otherwise 
                             // GPF-potential for 
                             // very high numbers
      do case             
         case nNumLeft = 0
              RETURN "null"
         case nNumLeft < 0
              cResult  = "minus "
              nNumLeft = -nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale = int( log10( nNumLeft ) / 3 )
         nGroup = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         if nGroup #1 or nScale < 1
            cResult += CLASS::Thou2GermanWords( nGroup ) 
         else 
            cResult+=iif(nScale<2,"ein","eine ")
         endif
         if nScale > 0
            if ngroup<2
               cresult+=trim( substr( "tausend     Million     Milliarde   Billion     Billiarde   Trillion ",;
                        nScale * 12 - 11, 12 ) )+iif(nscale>=2," ","")
            else
               cresult+=trim( substr( "tausend      Millionen   Milliarden  Billionen   Billiarden  Trillionen ",;
                        nScale * 12 - 11, 12 ) )+iif(nscale>=2," ","")
            endif
         endif
      enddo           

   RETURN trim(cResult)
   // -- EoM: Num2GermanWords()

   FUNCTION Thou2FrenchWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Jean-Pierre Martel
       Date........: 04/11/2005
       Notes.......: Converts a positive integer less than 1000 to
                     a string of characters.
       Written for.: dBASE PLUS
       Rev. History: Based on Num2Words by Jay Parsons.
                     Modified to display French words
       Calls.......: None
       Usage.......: convert.Thou2FrenchWords(<nNum>)
       Example.....: ? _app.Convert.Thou2FrenchWords( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local cUnits, cTens, nN, cResult, nHundreds, cHundreds
      cUnits = "un      deux    " ;
             + "trois   quatre  " ;
             + "cinq    six     " ;
             + "sept    huit    " ;
             + "neuf    dix     " ;
             + "onze    douze   " ;
             + "treize  quatorze" ;
             + "quinze  seize   " ;
             + "dix-septdix-huit" ;
             + "dix-neuf"
      cTens = "dix             vingt           ";
            + "trente          quarante        ";
            + "cinquante       soixante        ";
            + "soixante-dix    quatre-vingt    ";
            + "quatre-vingt-dix"
      nN = mod( int( nNum ), 100 )
      nHundreds = int(nNum/100)
      if nHundreds > 1
         cHundreds = cUnits.substring((nHundreds-1)*8, nHundreds*8).rightTrim() + " cents "
      elseif nHundreds = 1
         cHundreds = "cent "
      elseif nHundreds = 0
         cHundreds = ""
      endif
      do case
         case nNum = 0
            return("zro")
         case nNum = 100 and nN = 0
            return("cent")
         case nN = 0
            return( cHundreds.rightTrim() )
         case nN < 20
            return( cHundreds ;
            + cUnits.substring((nN-1)*8, nN*8).rightTrim() ;
            )
         case nN = 80
            return( cHundreds ;
              + "quatre-vingts";
              )
         case mod(nN,10) = 0
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() )
         case nN = 21 or nN = 31 or nN = 41 or nN = 51 or (nN > 60 and nN < 70)
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() ;
              + " et " ;
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
              )
         case (nN > 69 and nN < 80)
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-2)*16, (int(nN/10)-1)*16 ).rightTrim() ;
              + " et " ;
              + cUnits.substring( (mod(nN,10)+9)*8, (mod(nN,10)+10)*8 ).rightTrim() ;
              )
         case (nN > 90 and nN < 100)
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-2)*16, (int(nN/10)-1)*16 ).rightTrim() ;
              + " " ;
              + cUnits.substring( (mod(nN,10)+9)*8, (mod(nN,10)+10)*8 ).rightTrim() ;
              )
         otherwise
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() ;
              + "-" ;
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
              )
      endcase
      RETURN
   // -- EoM: Thou2FrenchWords()

   FUNCTION Thou2GermanWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Erik Moeller
       Date........: 06/22/1998
       Notes.......: Converts a positive integer less than 1000 to 
                     a German word.              
         				Based on Num2Words by Jay Parsons.
       Written for.: Visual dBASE 7
       Rev. History: 06/22/1998 - Original Release
                     06/23/1998 - Minor bug fix in "und" ...
                     06/25/1998 - Replaced German umlauts with 
                                  chr() calls
                     July 9, 2001 -- R. Wingerter noted
                                  chr() calls were for ASCII
                                  values, and replaced with ANSI
                                  values. If you are using a DOS
                                  language driver, use:
                                     Thou2GermanWordsASC instead.
                     July 12, 2001 -- R. Wingerter modified
                                  to allow use with both ANSI or
                                  DOS (ASCII) language drivers,
                                  obviating the need for the
                                     Thou2GermanWordsASC routine.
                                  Changed the error message to avoid
                                  a-Umlaut.
       Calls.......: None
       Usage.......: convert.Thou2GermanWords(<nNum>)
       Example.....: ? _app.Convert.Thou2GermanWords( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local cUnits, cTens, nN, frontstr, bottomstr, ec
      local coUml, cuUml, cSzlig
      if nnum >= 1000 or nnum < 0
         return "Nur Zahlen von 0 bis 999 erlaubt"	// changed message text
      endif

      // ********************************************************** //		
      // Added the following code to supply values for special characters
      // depending on global language driver. 
      // Could have used an array, but this may be easier to read
      // (the variable names are similar to the HTML named entity syntax): 
		// =&ouml, =&uuml,   = &szlig,  (we do not need = &auml)
      // Assumes that ANSI-Driver names always start with "DBWIN" 
      // ********************************************************** //      		
		if left(_app.lDriver,5) == "DBWIN" 		  	
			coUml=chr(246); cuUml=chr(252); cSzlig=chr(223)
      else
			coUml=chr(148); cuUml=chr(129); cSzlig=chr(225)
		endif
      // ********************************************************** //      		
            		
      frontstr=""
      bottomstr=""
      ec=0

      cUnits = "eins     zwei     " ;
             + "drei     vier     " ;
             + "f"+cuUml+"nf     sechs    " ;
             + "sieben   acht     " ;
             + "neun     zehn     " ;
             + "elf      zw"+coUml+"lf    " ;
             + "dreizehn vierzehn " ;
             + "f"+cuUml+"nfzehn sechzehn " ;
             + "siebzehn achtzehn " ;
             + "neunzehn "
      cTens = "zwan drei vier f"+cuUml+"nf sech sieb acht neun"
      nN = int( nNum )
      if nN = 0
         RETURN "null"
      endif
      nResult = ""

      if nN > 99
         posB= int(nN / 100 ) * 9 - 8
         frontstr+=iif(posb=1 AND nn>10 , "ein" ,trim( substr( cUnits, posB, 9 ))) + "hundert"
         nN = mod( nN, 100 )
         if nN = 0			
            RETURN frontstr
         endif
         frontstr+="und"
      endif

      if nN > 19
         posA=int(nn / 10) * 5 -9

         bottomstr = trim( substr( cTens, ;
                     posA, 5 ) ) + iif(posA # 6,"zig",cSzLig+"ig")
         ec = nn
         nN = mod( nN, 10 )
         if nN = 0 AND nNum < 100
            RETURN bottomstr
         else
            if mod(nn,10) > 0
               bottomstr = "und"+bottomstr
            endif
         endif
      endif

      posB= nn*9-8
      bottomstr=iif(posb=1 and ec>10, "ein" ,trim( substr( cUnits, posB, 9 )))+bottomstr
	RETURN frontstr+bottomstr
   // -- EoM: Thou2GermanWords()

   FUNCTION Num2SpanishWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Gerardo Pasos
       Date........: 11/26/1999
       Notes.......: Limited to 999,999,999,999,999.
                     In Mexico after the number 999,999,999. The
                     next number is one thousand million NOT one
                     Billion as it is in USA.
                     Based on Num2Words by Jay Parsons.
       Written for.: dBASE 7.01
       Rev. History: 03/01/1992 - Original Release.
       Calls.......: convert.Thou2spanishWords()
       Usage.......: convert.Num2spanishWords(<nNum>)
       Example.....: ? _app.Convert.Num2spanishWords( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local nNumleft, nScale, nGroup, cResult

      nNumLeft = int( nNum )
      do case
         case abs( nNumLeft ) >= 10 ^ 15
              RETURN "Imposible convertir un nmero mayor "+;
                     "que 999,999,999,999,999"
         case nNumLeft == 0
              RETURN "cero"
         case nNumLeft < 0
              cResult = "menos "
              nNumLeft = -nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale   = int( log10( nNumLeft ) / 3 )
         nGroup   = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         if nGroup=1 and nNumleft<>0
            cResult += "un"
         else
         cResult += CLASS::Thou2spanishWords( nGroup )
         endif
         if nNumLeft>0 and right(cResult,3)="uno"
            cResult=substr(cResult,1,len(cResult)-1)
         endif
         if nScale > 0
            cResult += " " ;
              + trim( substr( "mil     millon  mil     billon trillon",;
                nScale * 8 - 7, 8 ) )
            if nScale>1 and nScale<>3 and nGroup>1
               cResult += "es"
            endif
            if nNumLeft > 0
               cResult += " "
            endif
         endif
      enddo

   RETURN cResult
   // -- EoM: Num2SpanishWords()
   
   FUNCTION Thou2SpanishWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Gerardo Pasos
       Date........: 11/26/1999
       Notes.......: Converts a positive integer less than 1000 to
                     a string of characters.
       Written for.: dBASE 7.01
       Rev. History: Based on Num2Words by Jay Parsons.
                     Modified to work for Mexico and display
                     Spanish words
       Calls.......: None
       Usage.......: convert.Thou2Words(<nNum>)
       Example.....: ? _app.Convert.Thou2Words( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local cUnits, cTens, nN, cResult
      cUnits = "un       dos      " ;
             + "tres     cuatro   " ;
             + "cinco    seis     " ;
             + "siete    ocho     " ;
             + "nueve    diez     " ;
             + "once     doce     " ;
             + "trece    catorce  " ;
             + "quince   "
      cTens = "diez     veint    treinta  cuarenta cincuentasesenta setenta  ochenta  noventa  "
      cThou = "cien         doscientos   trescientos  " ;
            + "cuatrocientosquinientos   seiscientos  " ;
            + "setecientos  ochocientos  novecientos"
      nN = int( nNum )
      if nN = 0
         RETURN "zero"
      endif
      nResult = ""
      if nNum > 99
         nResult = trim( substr(cThou, int(nNum / 100 ) * ;
                            13 - 12, 13 ) )
         if nNum<200 and mod( nN, 100 )<>0
            nResult +="to"
         endif
         nN = mod( nN, 100 )
         if nN = 0
            RETURN nResult
         else
            nResult += " "
         endif
      endif
      if nN > 15
         nResult += trim( substr( cTens, ;
                    int( nN / 10 ) * 9 - 8, 9 ) )
         if nN=20
            nResult += "e"
         endif
         if nN>20 and nN<30
            nResult += "i"
         endif
         n20=nN
         nN = mod( nN, 10 )
         if nN = 0
            RETURN nResult
         else
            if n20>20 and n20<30

            else
              nResult += " y "
            endif
         endif
      endif
      if nN=1
         nResult += "uno"
      else
         nResult += trim( substr( cUnits, nN * 9 - 8, 9 ) )
      endif
   RETURN nResult
   // -- EoM: Thou2SpanishWords()

   function Num2Word_ITA(nNum)
   /*
      -------------------------------------------------------------
       Programmer..: Maurizio S.
       Date........:  February 28, 2020
       Notes.......: Created by Maurizio when he discovered there
                        was no version of this for Italian.
                     Based on Num2Words by Jay Parsons.
       Written for.: dBASE Plus 11
       Rev. History: 03/01/1992 - Original Release.
       Calls.......: convert.Thou2ITAWords()
       Usage.......: convert.Num2Word_ITA(<nNum>)
       Example.....: ? _app.Convert.Num2Word_ITA( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */
   
   ** 
   local nNumleft, nScale, nGroup, cResult

      nNumLeft = int( nNum )
      do case
         case abs( nNumLeft ) >= 10 ^ 18
              return("conversione impossibile")
         case nNumLeft == 0
              return("zero")
         case nNumLeft < 0
              cResult = "meno "
              nNumLeft = - nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale   = int( log10( nNumLeft ) / 3 )
         nGroup   = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         if nGroup = 1 and nNumleft <> 0
            cResult += "un"
         else
         cResult += CLASS::Thou2ITAWords( nGroup )
         endif
         if nNumLeft > 0 and cResult.right(3) == "un"
            cResult =  cResult.substring( 0, cResult.length )
         endif
         if nScale > 0
            cResult += iif(cResult.right(3)=="un ", "", "" );
              + trim( substr( "mila        milioni     miliardi    triliardi   ",;
                nScale * 12 - 11, 12 ) )
            if nGroup # 1 or nScale < 1
               cResult += ""
            endif
            if nNumLeft > 0
               cResult += ""
            endif
         endif
      enddo
      if not empty(cResult)
         *- decimal after letters
         cResult += iif( mod(nNum,1)>0, "/"+right(str(mod(nNum,1),3,2),2) , "")
      endif

   return ( cResult )
   *-- eom: Num2Word_ITA()
   
   function Thou2ITAWords( nNum )
    /*
      -------------------------------------------------------------
       Programmer..: Maurizio S.
       Date........:  February 28, 2020
       Notes.......: Created by Maurizio when he discovered there
                        was no version of this for Italian.
       Written for.: dBASE Plus
       Rev. History: Based on Num2Words by Jay Parsons.
       Calls.......: None
       Usage.......: convert.Thou2ITAWords(<nNum>)
       Example.....: ? _app.Convert.Thou2ITAWords( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */
   **	
   local cUnits, cTens, nN, cResult, nHundreds, cHundreds
      cUnits = "uno         due         " ;
             + "tre         quattro     " ;
             + "cinque      sei         " ;
             + "sette       otto        " ;
             + "nove        dieci       " ;
             + "undici      dodici      " ;
             + "tredici     quattordici " ;
             + "quindici    sedici      " ;
             + "diciassette diciotto    " ;
             + "diciannove  "
      cTens  = "dieci       vent        ";
             + "trent       quarant     ";
             + "cinquant    sessant     ";
             + "settant     ottant      ";
             + "novant      "
      nN = mod( int( nNum ), 100 )
      nHundreds = int(nNum/100)
      if nHundreds > 1
         cHundreds = cUnits.substring((nHundreds-1)*12, nHundreds*12).rightTrim() + "cento"
      elseif nHundreds = 1
         cHundreds = "cento"
      elseif nHundreds = 0
         cHundreds = ""
      endif
      do case
         case nNum = 0
            return("zero")
         case nNum = 100 and nN = 0
            return("cento")
         case nN = 0
            return( cHundreds.rightTrim() )
         case nN < 20
            return( cHundreds ;
            + cUnits.substring((nN-1)*12, nN*12).rightTrim() ;
            )
         case mod(nN,10) = 0
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*12, int(nN/10)*12 ).rightTrim() )
         case not ( nN = 21 or nN = 31 or nN = 41 or nN = 51 or  nN = 61 or  nN = 71 or  nN = 81 or  nN = 91 ;
                 or nN = 28 or nN = 38 or nN = 48 or nN = 58 or  nN = 68 or  nN = 78 or  nN = 88 or  nN = 98 )
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*12, int(nN/10)*12-1 ).rightTrim() ;
              + iif( int(nN/10)=2, "i","a" ) ;
              + cUnits.substring( (mod(nN,10)-1)*12, mod(nN,10)*12 ).rightTrim() ;
              )		
         otherwise
            return( cHundreds ;
              + cTens.substring( (int(nN/10)-1)*12, int(nN/10)*12 ).rightTrim() ;
              + "" ;
              + cUnits.substring( (mod(nN,10)-1)*12, mod(nN,10)*12 ).rightTrim() ;
              )
      endcase
   return
   // -- EoM: Thou2ITAWords()
   
   FUNCTION Num2WordsBE( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Gaetano De Luisi
       Date........: Feb 2021
       Notes.......: Converts integer numbers to Belgian word 
                     strings. Limited to numbers from -(10^18) 
                     to 10^19. 
         				Based on Num2Words by Jay Parsons.
       Written for.: Visual dBASE PLUS
       Rev. History: Feb 2021
       Calls.......: convert.Thou2BEWords()
       Usage.......: convert.Num2BEWords(<nNum>)
       Example.....: ? _app.Convert.Num2BEWords( 4321568357 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local nNumleft, nScale, nGroup, cResult

      nNumLeft = int( nNum )
      do case
         case abs( nNumLeft ) >= 10 ^ 18
              return("Impossible convertir un nombre plus grand "+;
                     "que 999,999,999,999,999,999")
         case nNumLeft == 0
              return("zro")
         case nNumLeft < 0
              cResult = "moins "
              nNumLeft = - nNumLeft
         otherwise
              cResult = ""
      endcase
      do while nNumLeft > 0
         nScale   = int( log10( nNumLeft ) / 3 )
         nGroup   = int( nNumLeft / 10 ^ ( 3 * nScale ) )
         nNumLeft = mod( nNumLeft, 10 ^ ( 3 * nScale ) )
         if nGroup = 1 and nScale = 1
            cResult += ""
         else
         cResult += CLASS::Thou2BEWords( nGroup )
         endif
         if nNumLeft > 0 and cResult.right(2) == "un"
            cResult =  cResult.substring( 0, cResult.length )
         endif
         if nScale = 1 and nGroup=1
				cResult +=  trim( substr( "mille       million     milliard    billion     billiard    ",;
                nScale * 12 - 11, 12 ))+IIF(nNumLeft>0,"-","")
			elseif nScale > 0
				// "mille is never preceded by "un" when at start of number
				// we say "thousand" and not "one thousand" in French
            cResult +="-"+ trim( substr( "mille       million     milliard    billion     billiard    ",;
                nScale * 12 - 11, 12 ) )
            if nGroup # 1 or nScale < 1					
               if "mille"$cResult
						//"mille" is invariable as it is already a plural - don't add an "s" for plural						
					else
						cResult+="s"
					endif
            endif
            if nNumLeft > 0
               cResult += "-"
            endif			
         endif
      enddo

   RETURN cResult
   // end of method: Num2WordsBE

   FUNCTION Thou2BEWords( nNum )
   /*
      -------------------------------------------------------------
       Programmer..: Gaetano De Luisi
       Date........: Feb 2021
       Notes.......: Converts a positive integer less than 1000 to
                     a string of characters.
       Written for.: dBASE PLUS
       Rev. History: Based on Num2Words by Jay Parsons.
                     Modified to display Belgian words
       Calls.......: None
       Usage.......: convert.Thou2BEWords(<nNum>)
       Example.....: ? _app.Convert.Thou2BEWords( 834 )
       Returns.....: Character string equivalent
       Parameters..: nNum = numeric integer to convert
      -------------------------------------------------------------
   */

      local cUnits, cTens, nN, cResult, nHundreds, cHundreds
      cUnits = "un      deux    " ;
             + "trois   quatre  " ;
             + "cinq    six     " ;
             + "sept    huit    " ;
             + "neuf    dix     " ;
             + "onze    douze   " ;
             + "treize  quatorze" ;
             + "quinze  seize   " ;
             + "dix-septdix-huit" ;
             + "dix-neuf"
      cTens = "dix             vingt           ";
            + "trente          quarante        ";
            + "cinquante       soixante        ";
            + "septante        quatre-vingt    ";
            + "nonante        "
      nN = mod( int( nNum ), 100 )
      nHundreds = int(nNum/100)
      if nHundreds > 1 and (nNum-(nHundreds*100))=0
         cHundreds = cUnits.substring((nHundreds-1)*8, nHundreds*8).rightTrim() + "-cents"
		elseif nHundreds > 1 and (nNum-(nHundreds*100))>0
			cHundreds = cUnits.substring((nHundreds-1)*8, nHundreds*8).rightTrim() + "-cent"		
      elseif nHundreds = 1 
        if nNum > 199 
				cHundreds = "-cent"			
			else
				cHundreds = "cent"
			endif
      elseif nHundreds = 0
         cHundreds = ""
      endif
      do case
         case nNum = 0
            return("zro")
         case nNum = 100 and nN = 0
            return("cent")
         case nN = 0
            return( cHundreds.rightTrim() )
         case nN < 20
            return ( IIF(nHundreds=0,"",cHundreds+"-");
            + cUnits.substring((nN-1)*8, nN*8).rightTrim() ;
            )
         case nN = 80
            return( IIF(nHundreds=0,"",cHundreds+"-") ;
              + "quatre-vingts";
              )
         case mod(nN,10) = 0
            return( IIF(nHundreds=0,"",cHundreds +"-");
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() )
			case nN = 21 or nN = 31 or nN = 41 or nN = 51 or nN = 61 or ;
						nN = 71 or nN = 91
            return( IIF(nHundreds=0,"",cHundreds+"-") ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() ;
              + "-et-" ;
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
              )
         case nN > 60 and nN < 70
            return( IIF(nHundreds=0,"",cHundreds+"-") ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() ;
              + "-" ;
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
              )
         case (nN > 69 and nN < 80)
            return( IIF(nHundreds=0,"",cHundreds+"-") ;
              + cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() ;
              + "-" ;
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
              )

         otherwise
            return( IIF(nHundreds=0,"",cHundreds+ "-" );
              +IIF(cTens=0,"",cTens.substring( (int(nN/10)-1)*16, int(nN/10)*16 ).rightTrim() );
              + "-";
              + cUnits.substring( (mod(nN,10)-1)*8, mod(nN,10)*8 ).rightTrim() ;
             )
					  
      endcase
   RETURN
   // -- EoM: Thou2WordsBE()

ENDCLASS
// -- End of Class: Convert

/*
   -----------------------------------------------------------------------
   End of File: CONVERT.CC
   ----------------------------------------------------------------------- 
*/
