class FileAttrib
/*
   ----------------------------------------------------------------
   Description:
 
    This class contains two methods -- one allows the developer
    to set file attributes the other returns the current
    attribute setting for a file. This is done using
    API calls, and is based on some messages by Steve Tipton in
    the Visual dBASE newsgroups .

    This is not subclassed from the file class, due to the
    fact that if you attempt to use a setAttribute call on
    a file that you currently have open using the file class
    you could easily screw yourself up. While some folk may
    claim that this would be programmer error, I'm not setting
    this up to make it any easier to shoot yourself in the foot.
     
    Programmers: 
 
       Ken Mayer (based on contributions by Steve Tipton)
 
    History: 
 
       11/11/1998 -- original
       February, 2002 -- Rich (at Traker International)
                         presented a new version of GetAttrib
                         that handles some attributes that
                         were added to the operating system
                         in later OSes, such as "Compressed",
                         and such. In addition, his version
                         of this method returns multiple
                         attributes.
       June 25, 2003 -- Minor changes based on suggestions by
                        Jim Sare, who pointed out this web page
                        from Microsoft:
      http://msdn.microsoft.com/library/default.asp?url=/library/en-us/fileio/base/setfileattributes.asp
 
    Instantiation: 
 
       set procedure to :dUFLP:fileattr.cc additive
       f = new FileAttrib()
 
    Methods:
       More details on each of these is contained in the 
       commentary for each method ...
 
       SetAttrib -- sets file attributes

       GetAttrib -- returns current attribute

 
    Properties:
 
       aAttr -- a two-dimensional array that stores
                the constants used for the api calls
   ----------------------------------------------------------------
*/

   // Build attribute array ... 
   // February, 2002: added Directory and Compressed options
   this.aAttr = new array( 8, 2 )
   this.aAttr[1,1] = "RO" // Read Only
   this.aAttr[1,2] = 1
   this.aAttr[2,1] = "H"  // Hidden
   this.aAttr[2,2] = 2
   this.aAttr[3,1] = "S"  // System
   this.aAttr[3,2] = 4  
   this.aAttr[4,1] = "D"  // Directory
   this.aAttr[4,2] = 16
   this.aAttr[5,1] = "A"  // Archive
   this.aAttr[5,2] = 32
   this.aAttr[6,1] = "N"  // Normal
   this.aAttr[6,2] = 128
   this.aAttr[7,1] = "T"  // Temporary
   this.aAttr[7,2] = 256
   this.aAttr[8,1] = "C"  // Compressed
   this.aAttr[8,2] = 2048 

   /*
      -------------------------------------------------------------
      This method of the FileAttrib class
      requires two parameters, a filename 
      (which can include a path), and 
      the attribute (from the list above) --
      use the 1 or two letter version ...

      Example: 
         f.SetAttrib( "c:\myfile", "H" ) // make it hidden

      If you wished to set the attributes of a set
      of files in a directory, you could do 
      something like:

      aMyTables = new array()
      aMyTables.dir( "*.DBF" )
      for nRow = 1 to aTables.subscript( aMyTables.size, 1 )
          f.SetAttrib( aMyTables[ nRow, 1 ], "RO" )
      next
      -------------------------------------------------------------
   */
   function SetAttrib( cFile, cAttrib )

      // Just in case ...:
      cAttrib = upper( cAttrib )

      // Check for empty parameters, file not existing,
      // or attribute not in list ...
      if ( empty( cFile ) or empty( cAttrib ) ) OR ;
         ( not file( cFile ) ) OR ;
         ( this.aAttr.scan( cAttrib ) == 0 )
         return null
      endif
      
      // prototype the API call:
      if type( "SetFileAttributes" ) # "FP"
         extern cLogical SetFileAttributes( cString, cULong );
                kernel32 from "SetFileAttributesA"
      endif
      // get the correct value ...
      nElement = this.aAttr.scan( cAttrib )
      nRow     = this.aAttr.subscript( nElement, 1 )

   return SetFileAttributes( cFile, this.aAttr[nRow,2] )

   function GetAttrib( cFile )
   /*
      -------------------------------------------------------------
      Programmers: Rich (?) 
                   Traker International www.autotraker.com
      Date.......: February, 2002
      Updates....: February, 2002 -- Ken Mayer
                   Some minor changes to make this more inline
                   with dUFLP standards, and to add commas
                   between attributes. Changed read-only
                   to return "RO", instead of just "R", like
                   the original version did.
      Usage......: f.GetAttrib( "c:\autoexec.bat" )
      Notes......: This version returns values that were not
                   available on NTFS/Win 2K, etc. It should
                   work the same as the previous version of this
                   code, but it doesn't use the array (currently)
                   and it returns multiple attributes, rather than
                   just a single attribute like the previous
                   version did.
      -------------------------------------------------------------
   */

      local nAttribute, cAttr, cReturn
      cAttr = ""
      cReturn = ""
      #DEFINE FILE_ATTRIBUTE_READONLY      1
      #DEFINE FILE_ATTRIBUTE_HIDDEN        2
      #DEFINE FILE_ATTRIBUTE_SYSTEM        4
      #DEFINE FILE_ATTRIBUTE_DIRECTORY    16
      #DEFINE FILE_ATTRIBUTE_ARCHIVE      32
      #DEFINE FILE_ATTRIBUTE_NORMAL      128
      #DEFINE FILE_ATTRIBUTE_TEMPORARY   256
      #DEFINE FILE_ATTRIBUTE_COMPRESSED 2048
      // suggested by Jim Sare, but not sure what to do with it
      // from here (Ken):
      #define FILE_ATTRIBUTE_NOT_CONTENT_INDEXED 0x00002000
      if type( "GetFileAttributesA" ) # "FP"
         extern cInt GetFileAttributesA(cString) kernel32 ;
                     from "GetFileAttributesA"
      endif
      nAttribute = GetFileAttributesA(cFile)
      // Check for errors:
      if nAttribute == -1
         cReturn = "File/Directory does not exist"
      endif
      if nAttribute < -1
         cReturn = "Unknown error"
      endif

      // Check for each attribute:
      if bitand(nAttribute,FILE_ATTRIBUTE_READONLY) ==;
                FILE_ATTRIBUTE_READONLY
         cAttr+='RO'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_ARCHIVE) ==;
                FILE_ATTRIBUTE_ARCHIVE
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='A'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_SYSTEM) ==;
                FILE_ATTRIBUTE_SYSTEM
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='S'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_HIDDEN) ==;
                FILE_ATTRIBUTE_HIDDEN
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='H'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_DIRECTORY) ==;
                FILE_ATTRIBUTE_DIRECTORY
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='D'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_NORMAL) ==;
                FILE_ATTRIBUTE_NORMAL
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='N'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_TEMPORARY) ==;
                FILE_ATTRIBUTE_TEMPORARY
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='T'
      endif
      if bitand(nAttribute,FILE_ATTRIBUTE_COMPRESSED) ==;
                FILE_ATTRIBUTE_COMPRESSED
         if not empty( cAttr )
            cAttr+=", "
         endif
         cAttr+='C'
      endif
      // if cReturn is empty, no error message, replace
      // contents with cAttr:
      if empty( cReturn )
         cReturn = cAttr
      endif
   return cReturn

endclass
/*
   ----------------------------------------------------------------
   End of File: FILEATTR.CC
   ----------------------------------------------------------------
*/
