/*
   RunCode.prg

   Author: Peter Rorlick

   (Based on someone else's Interpreter class, modified by Pete)

   Note: There is a limit as to how long a string can be if it is to be
   macro-expanded (around 300 bytes, depending on the complexity).
   To get around this, you can "break it up" into multiple code-block/strings,
   by putting ... at the end of any line.

   I recommend that you put ... at the end of every 5th line or so.
   But be careful NOT to put a ... in the middle of a control statment, such
   as IF/ENDIF or DO/ENDDO.

   If you want your code block to end prematurely, set STOP=true or RETURN 'STOP'.
   Examples:
     *-------
     if <some condition>
       stop = true
       return nResult
     endif
     *-------
     if <some condition>
       return 'stop'
     endif

   Empty lines in your text file will be ignored, as well as any "comment"
   lines that begin with a *

   Example of usage:
    set alternate to test.prg
    set alternate on
    ? [for i = 1 to 4  ]
    ? [  if msgbox("Do you want to stop?", "iteration "+i, 4+32)==6' ]
    ? [    stop = true ]
    ? [    exit        ]
    ? [  endif         ]
    ? [next...         ]
    ? [* This line is a comment.     ]
    ? [msgbox("You said NO 4 times") ]
    close alternate
    set alternate to
    do runcode.prg with 'test.prg'

*/

param cFileToRun, p1, p2, p3, p4, p5
local x
x = new interpreter()
x.file = cFileToRun
return x.execute()
**********************************************************************
class Interpreter
 this.file = ""
 // This version allows program flow commands: IF/ENDIF/SCAN/FOR

 Procedure Execute
  TRY
   if not file(this.file)
     msgbox( "File "+"<"+This.File+"> "+"does not exist!",'Error',48)
     return
   endif
   this.nHandle = fopen(this.File,"R")
   this.aExec = new array()              // Create array to store commands.
   if this.nHandle>0                     // Verify that we successfully opened the file.
     do while not feof(this.nHandle)    // Loop through the lines of the file.
        this.cString = fgets(this.nHandle)  // Read next line into cString property
        if right(This.cString,1) = CHR(26)
          // Remove this unwanted char:
          this.cString = left(this.cString,len(this.cString)-1)
        endif
        if left(ltrim(this.cString),1)="*" or len(trim(this.cString))=0
         // this is a comment or an empty line
        else
           this.aExec.add(ltrim(trim(this.cString)))  // Add command to array
        endif
     enddo
     fclose(this.nHandle)
   else
     msgbox("Could not open file!",'Error',48)
     return
   endif

   if type("this.aExec")="U"
     msgbox("File has not been read!",'Error',48)
     return
   endif
   // Put all the stuff in a codeblock
   private cCodeBlock, bBlock, ReturnValue, aVars, stop
   stop = false
   aVars = new AssocArray()  // for use in code blocks that are broken into "pieces" by "...".
   // Example: Runcb('aVars["msg"]="hello"...'+chr(13)+chr(10)+'msgbox(aVars["msg"])')
   n = 1
   do while n <= alen(this.aExec)
     cCodeblock = "{"
     do while n <= alen(this.aExec)   // for each "piece" of the code block...
        cCodeblock += ";"+this.aExec[n]
        if right(this.aExec[n],3) == '...'   // Implies a break
          cCodeBlock = Left(cCodeBlock,Len(cCodeBlock)-3) // Remove the "..."
          n++
          exit
        endif
        n++
     enddo
     cCodeblock += "}"
     bBlock = &cCodeBlock.
     returnValue = bBlock()
     if stop or ( type('returnValue')=='C' and upper(returnValue)=='STOP' )
       exit
     endif
   enddo
   return ReturnValue
  Catch ( Exception e )
   MsgBox(cCodeblock,"There is an error in this CodeBlock",16)
  EndTry
 return

endclass
************************************************************************
/*
RunCb functon

Author: Peter Rorlick

Runs a code block that's stored in a string variable, using 
the Interpreter class.  This is useful for running code that is:
 - stored in a memo field
 - entered by the user in an Editor control
 -  or in a string that
is built programmatically.

Example of usage:
 local c, crlf
 crlf = chr(13) + chr(10)
 c = 'for i = 1 to 4'                                               + crlf +;
     '  if msgbox("Do you want to stop?", "iteration "+i, 4+32)==6' + crlf +;
     '    stop = true'                                              + crlf +;
     '    exit'                                                     + crlf +;
     '  endif'                                                      + crlf +;
     'next...'                                                      + crlf +;
     'msgbox("You said NO 4 times")'
 set proc to runcode.prg additive
 RunCb(c)

*/
function RunCb
  param cb,p1,p2,p3,p4,p5  // cb a string containing a quasi code block
  tmpcb = cb
  If not empty(trim(ltrim(RemChars( CHR(13)+CHR(10)+CHR(141),tmpcb))))
    Try
      if Empty(cb)
        return
      endif
      // Run the code block that is stored in (cb):
      local cTempFileName
      cTempFileName = funique( "T???????.TMP" )
      set alte to (cTempFileName)
      set alte on
      set memowidth to 255
      ? cb
      close alte
      do RunCode WITH cTempFileName, p1, p2, p3, p4, p5
      erase (cTempFileName)
    Catch ( Exception e )
       MsgBox(cb,"There is an error in this CodeBlock",16)
    EndTry
  EndIf
return
****************************************************************************
Function RemChars( cChars2Rem, cFromList )
  // Returns a string free of each specific character in the string cChars2Rem
  local i
  if empty(cChars2Rem) or empty(cFromList)
   return cFromList
  endif
  for i = 1 TO LEN(cChars2Rem)
   cFromList = strtran( cFromList, subs(cChars2Rem,i,1), '' )
  next
return cFromList
****************************************************************************
Function StrTran( cTargetString, cSearch, cReplace)
  // By Peter Rorlick
  // Like Clipper's built-in function.
  if argCount() < 3
    cReplace = ""   // Omiting the 3rd parameter means "remove all instances of the searh string".
  endif
  local cString, nAt
  cString = cTargetString
  nAt = AT( cSearch, cString )
  do while nAt > 0
    cString = stuff(cString, nAt, LEN(cSearch), cReplace)
    nAt = AT( cSearch, cString )
  enddo
return cString

