/*
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter

     Last update.: December 5, 2006. 

     Notes.......: Class for working with SQL-Select-Statements
     					 Parses a SQL-String passed as a parameter 
                   and stores its parts in custom properties. 

                   The class has been tested with simple sql strings.
                   JOINs should work too, but complex queries with 
                   subqueries, union etc. are not supported.  

                   getClause() extracts a part of the SQL string
                   setClause() can be used to change a part of the SQL string

     Written for.: dBase Plus
     
     Rev. History: 0.1 Dec 4, 2006
                   First version posted in dbase.deutsch

                   0.2 Dec 5, 2006
                   - Fixed bug found by Ivar B. Jessen:
                     Keywords are now recognized regardless of case.
                   - The case of the sql statement passed will be preserved. 
                   - Fixed problem with fieldnames like "where1"
                     by adding space to Keywords
                   - Shortened code of getClause()

     Dependency..: :dUFLP:stringEx.cc

     Usage.......: oRef = new sqlSelect(<cSQL>)

     Example.....: set proc to sqlSelect.cc additive
                   o = new sqlSelect("SELECT * FROM fish WHERE id > 1 ORDER BY name")
                   ? o.Select
                   ? o.From 
                   ? o.Where
                   ? o.OrderBy

                   o.setClause("SELECT","species, name")
                   ? o.Select

						 o.setClause("ORDER BY","species, name")	
                   ? o.OrderBy

 						 o.setClause("WHERE","name LIKE '%fish'")	
                   ? o.Where

                   ? o.String

     Parameter...: cSQL: SQL-Statement
     ------------------------------------------------------------------
*/

FUNCTION parseSQL(cstr)
	
   /*
   Usage sample:

   oSQL = parseSQL( [SELECT id, name, f.'length cm' FROM fish f WHERE f.'length cm' > 100 ORDER BY name ])

   */

   o = new sqlSelect(cStr)

   ? o.Select
   ? o.From
   ? o.Where
   ? o.GroupBy
   ? o.OrderBy
return o


// CLASS sqlSelect ( cString ) of StringEx( cString ) from ":dUFLP:stringEx.cc"
CLASS sqlSelect ( cString ) of String( cString ) 

	// Note space character at the end of each keyword
   this.aKeywords = {"SELECT ", "FROM ", "WHERE ", "GROUP BY ", "ORDER BY "}

   this.string = cString

   this.Select		= class::getClause("SELECT")
   this.From		= class::getClause("FROM")
   this.Where		= class::getClause("WHERE")
   this.GroupBy	= class::getClause("GROUP BY")
   this.OrderBy	= class::getClause("ORDER BY")
  
   FUNCTION getClause( cName )
   	// Extracts the clause starting with cName from string 
		local cName, cSQL, cNextKeyword
     	local cRetVal, nIndex, i

      // Add space to cName
      cName = cName.LeftTrim().RightTrim().toUpperCase()+" "
      cSQL  = this.toUpperCase()
      
      cRetVal = ''

      nIndex = this.aKeywords.scan(cName)

      if nIndex > 0 and cName $ cSQL

         // Default value: remaining string
         cRetVal = this.Substring(cSQL.IndexOf(cName), cSQL.Length)

			// Check cSQL for keywords following cName 
         for i = nIndex+1 to this.aKeywords.size
            cNextKeyword = this.aKeywords[i]
            if cNextKeyword $ cSQL
            	// Return String up to this KW
               cRetVal = this.substrFromTo(cName, cNextKeyword)
               exit
            endif
         next i
			// ? nIndex, cName, cRetVal
      endif
   return cRetVal.RightTrim()

   FUNCTION setClause( cName, cValue )
   	// Set a new value for cName, e.g.
      // ORDER BY, WHERE condition etc.
      local cString, nIndex, i
      local cName, cValue

   	cName 	= cName.LeftTrim().RightTrim().toUpperCase()
		cValue 	= cValue.LeftTrim().RightTrim()

      // Keywords have trailing space
      nIndex = this.aKeywords.scan(cName+" ")

      if nIndex = 0
      	return false
      endif

      if left(cValue.ToUpperCase(), len(cName)) == cName
	      cString = cValue
      else
      	cString = cName + " " + cValue
      endif

      do case
         case cName = "SELECT"
      		this.Select = cString
         case cName = "FROM"
      		this.From 	= cString
         case cName = "WHERE"
      		this.Where 	= cString
         case cName = "ORDER BY"
      		this.OrderBy = cString
         case cName = "GROUP BY"
      		this.GroupBy = cString
      endcase
      class::updateString()
   return true

   FUNCTION substrFromTo( cFrom, cTo )
      // Extract substring between cFrom and cTo
      // using first occurrence of cTo.
      // Method is case insensitive, but will preserve
      // the case of the original SQL string.  
    	// Auxiliary method.

      local cRetVal, cStr, nFrom, nTo

      cStr  = this.toUpperCase()
      nFrom = cStr.IndexOf(cFrom.toUppercase()) 
      nTo 	= cStr.IndexOf(cTo.toUpperCase())

      if nFrom >= 0 and nTo > nFrom
      	// Return substring of *original* string.
         // This way we preserve the case of the original statement
        	cRetVal = this.Substring(nFrom, nTo)
      else
         cRetVal = ""
      endif
   return cRetVal

   FUNCTION updateString( NULL )
      // Update string from its properties
     	// Auxiliary method.
   	this.string = this.Select  + ' ' ;
                + this.From 	 + iif(empty(this.Where  ), '', ' ') ;
                + this.Where   + iif(empty(this.GroupBy), '', ' ') ;
                + this.GroupBy + iif(empty(this.OrderBy), '', ' ') ;
                + this.OrderBy
   return this.string

ENDCLASS