/* 
   VARSTORE.CC
   Programmer: James Bellew
   Date......: 06/19/20000

   A custom control to store variable values in a database
   for later retrieval. This is similar to using an .INI file
   I guess. However, by definition, an .INI file is for just
   for initialisation. This control can be used anywhere
   throughout your application.

   I use it to store:
   	Window sizes and positions
      Index names and last position in index
      Label formatting sizes

    ...basically any application variable that needs to be stored
    after exit.

   This custom control will automatically create the VARSTORE
   table to hold the variables.

   Set any default values for a variable in the VARSTOREDEFS.H file.
   You can also state which variables are to be stored independently
   for each user.

   If a variable is not found in the database for a read,
   the default value will be returned.


   The following lines need to be placed somewhere in your 
   application such that they are executed before you try to
   STORE or RETRIEVE any variables:

		set procedure to varstore.cc additive
      _app.oVarStore = new varstore()

	The following lines need to be placed in any program/class/etc.
   which needs to store or retrieve a variable:

		#include "varstore.h"

	The following lines need to be coded before exiting your
   application to free up memory:

     	_app.oVarStore.release()


   To store a variable, use:
   	STORE(value, name)

      e.g.

      STORE(form.top, "FORMTOP")

   To retrieve a variable, use:

		variable = RETRIEVE(name)


      e.g.

      form.top = RETRIEVE("FORMTOP")

    
*/

*****************************************************
CLASS VARSTORE
*****************************************************
	this.visible = false

	if not file('VARSTORE.DBF')
		create table VARSTORE ( ;
		   CNTLNAME	 	character(20),;
		   CNTLVALUE 	character(254),;
		   CNTLTYPE	 	character(1) )
		create index CNTLNAME on VARSTORE (CNTLNAME)
	endif

   this.DATAMODREF1 = new DATAMODREF()
  	this.DATAMODREF1.parent = this
   with (this.DATAMODREF1)
      filename = "varstore.dmd"
      dataModClass = "varstoreDataModule"
      share = 1
      active = true
      left = 12
      top = 0.5
   endwith

   this.rowset = this.DATAMODREF1.ref.varstore1.rowset
      

** Store a variable, with its value, to the databasee
   Function StoreVariable
   parameters cntlvalue, cntlname

      cntlname = this.uservar(cntlname)
      if this.rowset.findkey(cntlname + new string().replicate(" ", 20 - len(cntlname)))
      else
   	   this.rowset.beginappend()
   	   this.rowset.fields['CNTLNAME'].value = cntlname
      endif
      DO CASE
         CASE Type("cntlvalue") = "N"
            this.rowset.fields['CNTLVALUE'].value = str(CNTLVALUE,8,2)
            this.rowset.fields['CNTLTYPE'].value = "N"
         CASE Type("cntlvalue") = "C"
            this.rowset.fields['CNTLVALUE'].value = CNTLVALUE
            this.rowset.fields['CNTLTYPE'].value = "C"
         CASE Type("cntlvalue") = "D"
            this.rowset.fields['CNTLVALUE'].value = dtoc(CNTLVALUE)
            this.rowset.fields['CNTLTYPE'].value = "D"
         CASE Type("cntlvalue") = "L"
            if cntlvalue = .T.
               this.rowset.fields['CNTLVALUE'].value = "T"
            else
               this.rowset.fields['CNTLVALUE'].value =  "F"
            endif
            this.rowset.fields['CNTLTYPE'].value = "L"
         OTHERWISE
            if 6 = MsgBox("An invalid variable type was passed to the VARSTORE module." ,;
               "Error" , 16) && 0 + 16
            endif
      ENDCASE
      this.rowset.save()

   
** Retrieve a variable from the table, or from the default
** value list if not on table yet
   Function RetrieveVariable
   parameters cntlname
	   #include "../varstoredefs.h"
		basiccntlname = cntlname
      cntlname = this.uservar(cntlname)
      if this.rowset.findkey(cntlname + new string().replicate(" ", 20 - len(cntlname)))
         DO CASE
            CASE this.rowset.fields['CNTLTYPE'].value = "N"
               return_val = val(this.rowset.fields['CNTLVALUE'].value)
            CASE this.rowset.fields['CNTLTYPE'].value = "C"
               return_val = rtrim(this.rowset.fields['CNTLVALUE'].value)
            CASE this.rowset.fields['CNTLTYPE'].value = "D"
               return_val = ctod(this.rowset.fields['CNTLVALUE'].value)
            CASE this.rowset.fields['CNTLTYPE'].value = "L"
               if this.rowset.fields['CNTLVALUE'].value  = "T"
                  return_val = true
               else
                  return_val = false
               endif
            OTHERWISE
               if 6 = MsgBox("The VARSTORE module found an invalid variable type on the varstore database." +;
                  CHR(13) + "Variable name: " + NAME +;
                  CHR(13) + "Variable type: " + TYPE ,;                  
                  "Error" , 16) && 0 + 16
               endif
               return_val = .F.
         ENDCASE
      else
         temp = "DEFAULT_" + basiccntlname
         return_val = &temp
      endif
      this.storevariable(return_val, cntlname)
      return return_val


** determine whether this variable should be stored/retrieved
** at a user level, or application level
   Procedure UserVar(cntlname)
	   #include "../varstoredefs.h"
   	usercntlname = "USER_" + cntlname
   	if type(usercntlname) = 'U'
      	return cntlname
      endif
      if &usercntlname = true
      	return id() + cntlname
      endif
      return cntlname

   Procedure Release
   	this.DATAMODREF1.active = false
      release object this
      
      
ENDCLASS